
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: AcceptQuest, C_GossipInfo, CompleteQuest, GetNumQuestChoices, GetQuestReward, IsQuestCompletable, ReadyCheckFrameYesButton

local IsShiftKeyDown = IsShiftKeyDown
local next = next
local UnitIsDND = UnitIsDND

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Automation Interact")

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

M.Event = {
	GossipShow = "GOSSIP_SHOW",
	QuestComplete = "QUEST_COMPLETE",
	QuestDetail = "QUEST_DETAIL",
	QuestGreeting = "QUEST_GREETING",
	QuestProgress = "QUEST_PROGRESS",
	ReadyCheck = "READY_CHECK",
}

M.EventEnabled = {}

M.EventFrame = {
	[M.Event.GossipShow] = GossipFrame,
	[M.Event.QuestComplete] = QuestFrameRewardPanel,
	[M.Event.QuestDetail] = QuestFrameDetailPanel,
	[M.Event.QuestGreeting] = QuestFrameGreetingPanel,
	[M.Event.QuestProgress] = QuestFrameProgressPanel,
	[M.Event.ReadyCheck] = ReadyCheckFrame,
}

function M.GetCompletedQuests()
	local completedQuests = {}

	for questIndex = 1, GetNumActiveQuests() do
		if select(2, GetActiveTitle(questIndex)) then
			table.insert(completedQuests, questIndex)
		end
	end

	return completedQuests
end

function M.GetGossipCompletedQuests()
	return A.table.filter(C_GossipInfo.GetActiveQuests(), function (q) return q.isComplete end, true)
end

function M.GetGossipOptions()
	local gossipOptions = C_GossipInfo.GetOptions()
	return #gossipOptions == 1 and {gossipOptions[1]} or A.table.filter(gossipOptions, function (o) return o.selectOptionWhenOnlyOption end, true)
end

function M.IsActive()
	return UnitIsDND("player") ~= M.MODIFIER_KEY_FUNCTION()
end

function M:OnEnable()
	self:RegisterEvent("MODIFIER_STATE_CHANGED", "UpdateEvents")
	self:RegisterEvent("PLAYER_FLAGS_CHANGED", "UpdateEvents")
	self:RegisterMessage("SIXUI_CONFIG_CHANGED")
	self:RegisterMessage("SIXUI_LAYOUT_CHANGED", "Update")
	self:RegisterMessage("SIXUI_PROFILE_CHANGED", "Update")
	self:Update()
end

function M:Update()
	self:UpdateModifierKey()
	self:UpdateEventEnabled()
	self:UpdateEvents()
end

function M:UpdateEventEnabled()
	Mixin(self.EventEnabled, {
		[M.Event.GossipShow] = self.config.acceptQuests or self.config.chooseGossipOptions or self.config.completeQuests,
		[M.Event.QuestComplete] = self.config.completeQuests,
		[M.Event.QuestDetail] = self.config.acceptQuests,
		[M.Event.QuestGreeting] = self.config.acceptQuests or self.config.completeQuests,
		[M.Event.QuestProgress] = self.config.completeQuests,
		[M.Event.ReadyCheck] = self.config.acceptReadyChecks,
	})
end

function M:UpdateEvents()
	local isActive = self.IsActive()

	for _, event in next, self.Event do
		if isActive and self.EventEnabled[event] then
			self:RegisterEvent(event)

			if self.EventFrame[event]:IsShown() then
				self[event](self)
			end
		else
			self:UnregisterEvent(event)
		end
	end
end

function M:UpdateModifierKey()
	self.MODIFIER_KEY_FUNCTION = C.Enum.ModifierKeyFunction[self.config.modifierKey]
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

-- Select the first quest or option from a gossip list.
function M:GOSSIP_SHOW()
	local availableQuests = self.config.acceptQuests and C_GossipInfo.GetAvailableQuests() or {}
	local completedQuests = self.config.completeQuests and self.GetGossipCompletedQuests() or {}
	local gossipOptions = self.config.chooseGossipOptions and self.GetGossipOptions() or {}

	if #availableQuests > 0 then
		C_GossipInfo.SelectAvailableQuest(availableQuests[1].questID)
	elseif #completedQuests > 0 then
		C_GossipInfo.SelectActiveQuest(completedQuests[1].questID)
	elseif #gossipOptions > 0 then
		C_GossipInfo.SelectOption(gossipOptions[1].gossipOptionID)
	end
end

-- Complete a quest if there is one or less rewards.
function M:QUEST_COMPLETE()
	local numQuestChoices = GetNumQuestChoices()

	if numQuestChoices <= 1 then
		GetQuestReward(numQuestChoices)
	end
end

-- Accept a quest.
function M:QUEST_DETAIL()
	AcceptQuest()
end

-- Select the first quest from a quest greeting.
function M:QUEST_GREETING()
	local availableQuests = self.config.acceptQuests and GetNumAvailableQuests() or 0
	local completedQuests = self.config.completeQuests and self.GetCompletedQuests() or 0

	if availableQuests > 0 then
		SelectAvailableQuest(1)
	elseif #completedQuests > 0 then
		SelectActiveQuest(completedQuests[1])
	end
end

-- Advance the quest completion dialog to the reward selection step.
function M:QUEST_PROGRESS()
	if IsQuestCompletable() then
		CompleteQuest()
	end
end

function M:SIXUI_CONFIG_CHANGED(event, module, config, key, ...)
	if module == self:GetName() then
		self:Update()
	end
end

-- Accept a ready check.
function M:READY_CHECK()
	ReadyCheckFrameYesButton:Click()
end
