
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: AuraButtonMixin, CreateFont, string

local BuffFrame = BuffFrame
local coroutine = coroutine
local DebuffFrame = DebuffFrame
local next = next
local unpack = unpack

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Buff Frame")

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

M.AuraType = {
	Buff = "Buff",
	DeadlyDebuff = "DeadlyDebuff",
	Debuff = "Debuff",
	TempEnchant = "TempEnchant",
}

M.AuraTypeIsBuff = {
	[M.AuraType.Buff] = true,
	[M.AuraType.TempEnchant] = true,
}

M.AuraTypeIsDebuff = {
	[M.AuraType.DeadlyDebuff] = true,
	[M.AuraType.Debuff] = true,
}

M.AuraFrame = {
	[M.AuraType.Buff] = BuffFrame.auraFrames,
	[M.AuraType.Debuff] = DebuffFrame.auraFrames,
}

function M.IterateAuras(auraType)
	return coroutine.wrap(function ()
		for _, auraFrames in next, not auraType and M.AuraFrame or {M.AuraFrame[auraType]} do
			for _, aura in next, auraFrames do
				if not aura.isAuraAnchor then
					coroutine.yield(aura)
				end
			end
		end
	end)
end

function M:OnEnable()
	self:UpdateFonts()
	self:CreateHook("SecureHook", AuraButtonMixin, "OnUpdate", "AuraButtonMixin_OnUpdate")
	self:CreateHook("SecureHook", BuffFrame, "UpdateAuraButtons", "BuffFrame_UpdateAuraButtons")
	self:CreateHook("SecureHook", BuffFrame, "UpdateGridLayout", "BuffFrame_UpdateAuraButtons")
	self:CreateHook("SecureHook", DebuffFrame, "UpdateAuraButtons", "DebuffButton_UpdateAuraButtons")
	self:CreateHook("SecureHook", DebuffFrame, "UpdateGridLayout", "DebuffButton_UpdateAuraButtons")
	self:RegisterMessage("SIXUI_CONFIG_CHANGED")
	self:RegisterMessage("SIXUI_PROFILE_CHANGED", "Update")
	self:StyleAllAuras()
end

function M:OnDisable()
	self:ResetAllAuras()
end

function M:Update()
	self:ResetAllAuras()
	self:StyleAllAuras()
end

function M:StyleAllAuras()
	for aura in self.IterateAuras() do
		self:StyleAura(aura)
	end
end

function M:ResetAllAuras()
	for aura in self.IterateAuras() do
		self:ResetAura(aura)
	end
end

function M:StyleAura(aura)
	if (not self.AuraType[aura.auraType]) then return end
	if (self.AuraTypeIsBuff[aura.auraType] and not self.config.skinBuffs) then return end
	if (self.AuraTypeIsDebuff[aura.auraType] and not self.config.skinDebuffs) then return end

	A:GetFrameConfig(aura):CreateBorder(C.BORDER_PADDING, C.BORDER_PADDING, C.BORDER_PADDING - 10, C.BORDER_PADDING)
	A:GetFrameConfig(aura.Count):ClearAllPoints():SetPoint("BOTTOMRIGHT", aura.Icon, "BOTTOMRIGHT", 0, 3):SetFontObject(self.TEXT_COUNT)
	A:GetFrameConfig(aura.DebuffBorder):SetShown(false)
	A:GetFrameConfig(aura.Duration):ClearAllPoints():SetPoint("TOP", aura.Icon, "BOTTOM", 0, -(C.BORDER_PADDING + 2)):SetFontObject(self.TEXT_DURATION)
	A:GetFrameConfig(aura.Icon):SetTexCoord(unpack(C.ZOOMED_TEX_COORDS[0.3]))
	A:GetFrameConfig(aura.TempEnchantBorder):SetShown(false)

	if self.AuraTypeIsDebuff[aura.auraType] then
		aura:GetBorder():SetColor(C.DEBUFF_TYPE_TO_COLOR[(aura.buttonInfo or {}).dispelName]:GetRGBA())
	end
end

function M:ResetAura(aura)
	A:GetFrameConfig(aura):ResetAll()
	A:GetFrameConfig(aura.Count):ResetAll()
	A:GetFrameConfig(aura.DebuffBorder):ResetAll():Run(function (f) f:SetShown(self.AuraTypeIsDebuff[aura.auraType]) end)
	A:GetFrameConfig(aura.Duration):ResetAll()
	A:GetFrameConfig(aura.Icon):ResetAll()
	A:GetFrameConfig(aura.TempEnchantBorder):ResetAll():Run(function (f) f:SetShown(aura.auraType == self.AuraType.TempEnchant) end)
end

function M:UpdateFonts()
	local fontFile, fontSize = C.TEXT_REGULAR:GetFont()
	self.TEXT_COUNT = self.TEXT_COUNT or CreateFont("SixUI Buff Count")
	self.TEXT_COUNT:SetFont(fontFile, fontSize, "OUTLINE")
	self.TEXT_DURATION = self.TEXT_DURATION or CreateFont("SixUI Buff Duration")
	self.TEXT_DURATION:SetFont(fontFile, fontSize, "")
	self.TEXT_DURATION:SetShadowColor(0, 0, 0, 1)
	self.TEXT_DURATION:SetShadowOffset(1, -1)
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function M:AuraButtonMixin_OnUpdate(aura)
	if not aura.isAuraAnchor then
		self:StyleAura(aura)
	end
end

function M:BuffFrame_UpdateAuraButtons()
	for aura in self.IterateAuras(self.AuraType.Buff) do
		self:StyleAura(aura)
	end
end

function M:DebuffButton_UpdateAuraButtons()
	for aura in self.IterateAuras(self.AuraType.Debuff) do
		self:StyleAura(aura)
	end
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:SIXUI_CONFIG_CHANGED(event, key, ...)
	local c, g, k = string.split(".", key)

	if key == self:GetName() then
		self:Update()
	elseif c == "color" and g == "debuff" then
		self:Update()
	elseif c == "font" then
		self:UpdateFonts()
	end
end
