
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: IsCombatLog, IsVoiceTranscription

local format = format
local string = string
local table = table
local tonumber = tonumber
local type = type

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Chat Abbreviations")

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

M.Abbreviation = {
	GUILD = "G",
	INSTANCE_CHAT = "I",
	INSTANCE_CHAT_LEADER = "IL",
	OFFICER = "O",
	PARTY = "P",
	PARTY_GUIDE = "PL",
	PARTY_LEADER = "PL",
	RAID = "R",
	RAID_LEADER = "RL",
}

M.Pattern = {
	FLAG_AFK = CHAT_FLAG_AFK .. "(|Hplayer)",
	FLAG_DND = CHAT_FLAG_DND .. "(|Hplayer)",
	RAID_WARNING = "^" .. CHAT_RAID_WARNING_GET:format("(.-|h)"),
	SAY = "^" .. CHAT_SAY_GET:format("(.-|h)"),
	TIMESTAMP = "^(%d-:%d-[:%d AMP]+) ",
	WHISPER_FROM = "^" .. CHAT_WHISPER_GET:format("(.-|h)"),
	WHISPER_TO = "^" .. CHAT_WHISPER_INFORM_GET:format("(.-|h)"),
	YELL = "^" .. CHAT_YELL_GET:format("(.-|h)"),
}

function M.AbbreviateChannel(channel)
	return format("|Hchannel:%s|h[%s]|h", channel, M.Abbreviation[channel] or channel:gsub("^.-:", ""))
end

function M.AbbreviateMessage(text, ...)
	local time = text:match(M.Pattern.TIMESTAMP)

	-- Remove the timestamp.
	text = text:gsub(M.Pattern.TIMESTAMP, "")

	-- Resize textures to match the font size.
	text = text:gsub("(|[AT]:?)([^:]+:[1-9][^|]*)(|[at])", M.ResizeTexture)

	-- Abbreviate message descriptors.
	text = text:gsub(M.Pattern.WHISPER_TO, "[W To] %1 ")
	text = text:gsub(M.Pattern.WHISPER_FROM, "[W From] %1 ")
	text = text:gsub(M.Pattern.SAY, "[S] %1 ")
	text = text:gsub(M.Pattern.YELL, "[Y] %1 ")
	text = text:gsub(M.Pattern.RAID_WARNING, "[RW] %1 ")

	-- Abbreviate channel names.
	text = text:gsub("^|Hchannel:(.-)|h.-|h", M.AbbreviateChannel)

	-- Strip the <Away> and <Busy> prefix from player names.
	text = text:gsub(M.Pattern.FLAG_AFK, "%1")
	text = text:gsub(M.Pattern.FLAG_DND, "%1")

	-- Strip the colon symbol from messages without a player name.
	text = text:gsub("^(|Hchannel.-|h.-|h) :", "%1")

	-- Restore the timestamp.
	if time then
		text = "[" .. time .. "] " .. text
	end

	return text, ...
end

function M.ResizeTexture(open, data, close)
	local args = {string.split(":", data, 4)}
	local height, width = tonumber(args[2]), tonumber(args[3])

	if height == width then
		args[2], args[3] = 0, 0
	else
		args[2], args[3] = 0, width and width / height or 0
	end

	return open .. table.concat(args, ":") .. close
end

function M:OnInitialize()
	self.IterateChatFrames = A:GetModule("Chat Frame").IterateChatFrames
end

function M:OnEnable()
	self:ConfigureAllChatFrames()
end

function M:ConfigureAllChatFrames()
	for chatFrame in self.IterateChatFrames() do
		self:ConfigureChatFrame(chatFrame)
	end
end

function M:ConfigureChatFrame(chatFrame)
	if not IsCombatLog(chatFrame) and not IsVoiceTranscription(chatFrame) then
		self:CreateHook("SecureHook", chatFrame, "AddMessage", "AbbreviateMessageByText")
		self:CreateHook("SecureHook", chatFrame, "BackFillMessage", "AbbreviateMessageByText")
		self:AbbreviateAllMessages(chatFrame)
	end
end

function M:AbbreviateAllMessages(chatFrame)
	chatFrame:TransformMessages(function (t) return type(t) == "string" end, self.AbbreviateMessage)
end

function M:AbbreviateMessageByText(chatFrame, text)
	if type(text) == "string" then
		chatFrame:TransformMessages(function (t) return t == text end, self.AbbreviateMessage)
	end
end
