
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: ConvertRGBtoColorString, FONT_COLOR_CODE_CLOSE, format, math, string, table, tonumber, UISpecialFrames, UNKNOWN, YELLOW_FONT_COLOR_CODE

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Chat Copy")
local Social = LibStub("SixSocial-1.0")

local CopyBox = SixUI_CopyBox
local ScrollFrame = CopyBox.scrollFrame
local EditBox = ScrollFrame.editBox

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

-- Balance color codes (|c...) and escape sequences (|r).
function M.BalanceColorCodes(inputString)
	local colorSeqCount, escapeSeqCount = 0, 0

	for colorSeq, escapeSeq in string.gmatch(inputString, "(|c%x%x%x%x%x%x%x%x)(|r?)") do
		colorSeqCount = colorSeqCount + (colorSeq and 1 or 0)
		escapeSeqCount = escapeSeqCount + (escapeSeq == "|r" and 1 or 0)
	end

	return inputString .. string.rep("|r", math.max(0, colorSeqCount - escapeSeqCount))
end

-- Format a BattleTag for display.
function M.FormatBattleTag(accountID)
	local battleTag = (Social:GetFriendByAccountID(tonumber(accountID)) or {}).battleTag
	return format("[%s]", string.split("#", battleTag or UNKNOWN))
end

function M.FormatIndefiniteArticle(word)
	local article = "a"

	if word:sub(1, 1):match("[aeiouAEIOU]") then
		article = "an"
	end

	return article .. " " .. word
end

-- Format a numeric value with the appropriate singular or plural text label.
function M.FormatPlural(value, singular, plural)
	return format("%s %s", value, value == "1" and singular or plural)
end

function M:OnInitialize()
	self.IterateChatFrames = A:GetModule("Chat Frame").IterateChatFrames
	table.insert(UISpecialFrames, "SixUI_CopyBox")
	EditBox:SetFontObject(C.TEXT_MEDIUM)
end

function M:OnEnable()
	self:RegisterMessage("SIXUI_COPY_LINES")
	self:ConfigureChatFrames()
end

function M:ConfigureChatFrames()
	for chatFrame in self.IterateChatFrames() do
		self:CreateHook("HookScript", chatFrame, "OnMouseDown", "ChatFrame_OnMouseDown")
	end
end

function M:CopyLines(lines)
	if CopyBox:IsShown() then
		CopyBox:Hide()
	else
		EditBox:SetText(table.concat(lines, "\n"))
		CopyBox:Show()
		RunNextFrame(function () self:ScrollToBottom() end)
	end
end

function M:GetLines(chatFrame)
	local lines = {}

	for index = 1, chatFrame:GetNumMessages() do
		local text, r, g, b = chatFrame:GetMessageInfo(index)
		local colorCode = r and g and b and ConvertRGBtoColorString({r = r, g = g, b = b}) or YELLOW_FONT_COLOR_CODE

		text = format("%s%s|r", colorCode, text:gsub(FONT_COLOR_CODE_CLOSE, FONT_COLOR_CODE_CLOSE .. colorCode))
		text = self.BalanceColorCodes(text)
		text = text:gsub(" ?|[AT]:?[^|]+|[at] ?", "")
		text = text:gsub("(%d+) |[47]([^:]+):([^;]+);", self.FormatPlural)
		text = text:gsub("|5([^ ]+)", self.FormatIndefiniteArticle)
		text = text:gsub("|HBNplayer:[^:]-:(%d+):.-|h.-|h", self.FormatBattleTag)
		text = text:gsub("|K[^|]+|k", "<" .. UNKNOWN .. ">")

		table.insert(lines, text)
	end

	return lines
end

function M:ScrollToBottom()
	ScrollFrame:SetVerticalScroll(ScrollFrame:GetVerticalScrollRange())
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:SIXUI_COPY_LINES(event, lines)
	self:CopyLines(lines)
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function M:ChatFrame_OnMouseDown(chatFrame, button)
	if button == "RightButton" then
		self:CopyLines(self:GetLines(chatFrame))
	end
end
