
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: C_Reputation, ChatFrame_AddMessageEventFilter, ChatFrame_RemoveMessageEventFilter, coroutine, GetGuildInfo, GUILD

local format = format
local next = next
local string = string

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Chat Faction")

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

M.EVENTS_WITH_FACTION = {
	"CHAT_MSG_COMBAT_FACTION_CHANGE",
}

M.FACTION_NAME_PATTERNS = {
	FACTION_STANDING_DECREASED_GENERIC:format("(.-)"),
	FACTION_STANDING_DECREASED_GENERIC_ACCOUNT_WIDE:format("(.-)"),
	FACTION_STANDING_INCREASED_GENERIC:format("(.-)"),
	FACTION_STANDING_INCREASED_GENERIC_ACCOUNT_WIDE:format("(.-)"),
}

function M.FilterFaction(self, event, message, ...)
	local name = M:ParseFactionName(message)

	if name then
		local link = format("|Hfaction:%s|h[%s]|h", name, name)
		local escapedLink = link:gsub("%%", "%%%%")
		local escapedName = name:gsub("[%(%)%.%%%+%-%*%?%[%^%$%]]", "%%%1")

		return false, message:gsub(escapedName, escapedLink), ...
	end

	return false, message, ...
end

function M.IterateFactionData()
	local numFactions = C_Reputation.GetNumFactions()

	return coroutine.wrap(function ()
		for factionIndex = 1, numFactions do
			local factionData = C_Reputation.GetFactionDataByIndex(factionIndex)

			if factionData and (not factionData.isHeader or factionData.isHeaderWithRep) then
				coroutine.yield(factionData)
			end

			if factionData and factionData.isCollapsed then
				C_Reputation.ExpandFactionHeader(factionIndex)
				numFactions = C_Reputation.GetNumFactions()
			end
		end
	end)
end

function M:OnEnable()
	self:CreateHook("SecureHook", "SetItemRef")
	self:AddMessageEventFilters()
end

function M:OnDisable()
	self:RemoveMessageEventFilters()
end

-- Add a message event filter for each event.
function M:AddMessageEventFilters()
	for _, event in next, self.EVENTS_WITH_FACTION do
		ChatFrame_AddMessageEventFilter(event, self.FilterFaction)
	end
end

-- Add a message event filter for each event.
function M:RemoveMessageEventFilters()
	for _, event in next, self.EVENTS_WITH_FACTION do
		ChatFrame_RemoveMessageEventFilter(event, self.FilterFaction)
	end
end

-- Get faction data by name.
function M:GetFactionDataByName(name)
	if name == GUILD then
		name = GetGuildInfo("player") or name
	end

	for factionData in self.IterateFactionData() do
		if factionData.name == name then
			return factionData
		end
	end
end

-- Parse a faction name from a message.
function M:ParseFactionName(message)
	for _, pattern in next, self.FACTION_NAME_PATTERNS do
		local name = message:match(pattern)

		if name then
			return name
		end
	end
end

-- Set the watched faction by name.
function M:SetWatchedFactionByName(name, silent)
	local factionData = self:GetFactionDataByName(name)

	if factionData and not factionData.isWatched then
		C_Reputation.SetWatchedFactionByID(factionData.factionID)

		if not silent then
			A.PrintF("You are now watching '%s'.", A.Highlight(name))
		end
	end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function M:SetItemRef(link)
	local type, value = string.split(":", link, 2)

	if type == "faction" then
		self:SetWatchedFactionByName(value)
	end
end
