
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: LibStub

local _G = _G
local InCombatLockdown = InCombatLockdown
local Mixin = Mixin
local next = next
local type = type
local unpack = unpack

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Core Registry")

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

M.frameConfigs = {}

function M:OnInitialize()
	self:CreateQueues()
	self:CreateHook("SecureHook", "CreateFrame")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
end

function M:CreateQueues()
	-- Create a queue for callbacks waiting for combat to end.
	self.CombatQueue = LibStub("SixQueue-1.0"):CreateQueue()

	-- Create a queue for callbacks waiting for a frame to exist.
	self.FrameQueue = LibStub("SixQueue-1.0"):CreateQueue()
end

function A:GetFrameConfig(key)
	if type(key) ~= "string" and type(key) ~= "table" then
		return Mixin({}, M.FrameConfigMixin)
	end

	if not M.frameConfigs[key] then
		M.frameConfigs[key] = Mixin({key = key}, M.FrameConfigMixin)
	end

	return M.frameConfigs[key]
end

function A:ProcessFrame(key)
	if key then
		M.FrameQueue:Process(key)
	end
end

function A:RunSecure(key, callback, ...)
	if InCombatLockdown() then
		M.CombatQueue:Add(key, callback, ...)
	else
		callback(...)
	end
end

-------------------------------------------------------------------------------
-- Frame Config Mixin
-------------------------------------------------------------------------------

M.FrameConfigMixin = {}

function M.FrameConfigMixin:CanRun(callback, dependencyFrame, ...)
	local frame = (type(self.key) == "table" and self.key) or (type(self.key) == "string" and _G[self.key])
	local dependency = (type(dependencyFrame) == "table" and dependencyFrame) or (type(dependencyFrame) == "string" and _G[dependencyFrame]) or nil

	self.frame = frame

	if not frame then
		-- The frame doesn't exist.
		if type(self.key) == "string" then
			M.FrameQueue:Add(self.key, callback, self, ...)
		end
	elseif dependencyFrame and not dependency then
		-- The dependency frame doesn't exist.
		if type(dependencyFrame) == "string" then
			M.FrameQueue:Add(dependencyFrame, callback, self, ...)
		end
	elseif frame:IsForbidden() then
		-- The frame is forbidden.
		return false
	elseif frame:IsProtected() and InCombatLockdown() and self.ShouldRunSecure[callback] then
		-- The frame is protected.
		M.CombatQueue:Add({}, callback, self, ...)
	else
		return true
	end
end

function M.FrameConfigMixin:ClearAllPoints()
	if self:CanRun(self.ClearAllPoints) then
		self.overridePoints = {}
		self:UpdatePoints()
	end

	return self
end

function M.FrameConfigMixin:CreateBorder(...)
	if self:CanRun(self.CreateBorder, nil, ...) then
		if not self.frame.GetBorder then
			self.originalBorder = false
			B:CreateBorder(self.frame, ...)
		end
	end

	return self
end

function M.FrameConfigMixin:DisableDrawLayer(drawLayer)
	if self:CanRun(self.DisableDrawLayer, nil, drawLayer) then
		if not self.disabledDrawLayers then
			self.disabledDrawLayers = {}
		end

		self.disabledDrawLayers[drawLayer] = true
		self.frame:DisableDrawLayer(drawLayer)
	end

	return self
end

function M.FrameConfigMixin:RemoveMaskTexture(mask)
	if self:CanRun(self.RemoveMaskTexture, nil, mask) then
		if not self.originalMaskTextures then
			self.originalMaskTextures = {}
		end

		self.originalMaskTextures[mask] = true
		self.frame:RemoveMaskTexture(mask)
	end

	return self
end

function M.FrameConfigMixin:Run(callback, ...)
	if self:CanRun(self.Run, nil, callback, ...) then
		callback(self.frame, ...)
	end

	return self
end

function M.FrameConfigMixin:RunSecure(callback, ...)
	if self:CanRun(self.Run, nil, callback, ...) then
		callback(self.frame, ...)
	end

	return self
end

function M.FrameConfigMixin:SetAllPoints(relativeRegion)
	if self:CanRun(self.SetAllPoints, relativeRegion, relativeRegion) then
		if not self.overridePoints then
			self.overridePoints = {}
		end

		self.overridePoints[#self.overridePoints] = {"SetAllPoints", relativeRegion}
		self:UpdatePoints()
	end

	return self
end

function M.FrameConfigMixin:SetAlpha(...)
	if self:CanRun(self.SetAlpha, nil, ...) then
		if not self.originalAlpha then
			self.originalAlpha = self.frame:GetAlpha()
		end

		self.frame:SetAlpha(...)
	end

	return self
end

function M.FrameConfigMixin:SetDrawLayer(...)
	if self:CanRun(self.SetDrawLayer, nil, ...) then
		if not self.originalDrawLayer then
			self.originalDrawLayer = self.frame:GetDrawLayer()
		end

		self.frame:SetDrawLayer(...)
	end

	return self
end

function M.FrameConfigMixin:SetFont(...)
	if self:CanRun(self.SetFont, nil, ...) then
		if not self.originalFont then
			self.originalFont = {self.frame:GetFont()}
		end

		self.frame:SetFont(...)
	end

	return self
end

function M.FrameConfigMixin:SetFontObject(...)
	if self:CanRun(self.SetFontObject, nil, ...) then
		if not self.originalFontObject then
			self.originalFontObject = self.frame:GetFontObject()
		end

		self.frame:SetFontObject(...)
	end

	return self
end

function M.FrameConfigMixin:SetParent(parent)
	if self:CanRun(self.SetParent, parent, parent) then
		if not self.originalParent then
			self.originalParent = self.frame:GetParent()
		end

		self.frame:SetParent(type(parent) == "table" and parent or _G[parent])
	end

	return self
end

function M.FrameConfigMixin:SetPoint(point, relativeTo, relativePoint, offsetX, offsetY)
	if self:CanRun(self.SetPoint, relativeTo, point, relativeTo, relativePoint, offsetX, offsetY) then
		if not self.overridePoints then
			self.overridePoints = {}
		end

		self.overridePoints[#self.overridePoints] = {"SetPoint", point, relativeTo, relativePoint, offsetX, offsetY}
		self:UpdatePoints()
	end

	return self
end

function M.FrameConfigMixin:SetScale(...)
	if self:CanRun(self.SetScale, nil, ...) then
		if not self.originalScale then
			self.originalScale = self.frame:GetScale()
		end

		self.frame:SetScale(...)
	end

	return self
end

function M.FrameConfigMixin:SetShown(shown)
	if self:CanRun(self.SetShown, nil, shown) then
		if self.originalIsShown == nil then
			self.originalIsShown = self.frame:IsShown()
		end

		if shown then
			M:Unhook(self.frame, "OnShow")
			M:CreateHook("HookScript", self.frame, "OnHide", "Frame_OnHide")
		else
			M:Unhook(self.frame, "OnHide")
			M:CreateHook("HookScript", self.frame, "OnShow", "Frame_OnShow")
		end

		self.frame:SetShown(shown)
	end

	return self
end

function M.FrameConfigMixin:SetTexCoord(...)
	if self:CanRun(self.SetTexCoord, nil, ...) then
		if not self.originalTexCoord then
			self.originalTexCoord = {self.frame:GetTexCoord()}
		end

		self.frame:SetTexCoord(...)
	end

	return self
end

function M.FrameConfigMixin:SetTexture(...)
	if self:CanRun(self.SetTexture, nil, ...) then
		if not self.originalTexture then
			self.originalTexture = {self.frame:GetTexture()}
		end

		self.frame:SetTexture(...)
	end

	return self
end

function M.FrameConfigMixin:UpdatePoints()
	if not self.updatePointMutex and self.overridePoints and self:CanRun(self.UpdatePoints) then
		self.updatePointMutex = true

		if not self.originalPoints then
			self.originalPoints = {}

			for i = 1, self.frame:GetNumPoints() do
				self.originalPoints[i] = {self.frame:GetPoint(i)}
			end
		end

		self.frame:ClearAllPoints()

		for _, point in next, self.overridePoints do
			self.frame[point[1]](self.frame, select(2, unpack(point)))
		end

		M:CreateHook("SecureHook", self.frame, "ClearAllPoints", function () self:UpdatePoints() end)
		M:CreateHook("SecureHook", self.frame, "SetAllPoints", function () self:UpdatePoints() end)
		M:CreateHook("SecureHook", self.frame, "SetPoint", function () self:UpdatePoints() end)

		self.updatePointMutex = nil
	end

	return self
end

function M.FrameConfigMixin:ResetAll()
	self:ResetAlpha()
	self:ResetBorder()
	self:ResetDrawLayers()
	self:ResetFont()
	self:ResetFontObject()
	self:ResetMaskTextures()
	self:ResetParent()
	self:ResetPoints()
	self:ResetScale()
	self:ResetShown()
	self:ResetTexCoord()
	self:ResetTexture()

	return self
end

function M.FrameConfigMixin:ResetAlpha()
	if self.originalAlpha and self:CanRun(self.ResetAlpha) then
		self.frame:SetAlpha(self.originalAlpha)
		self.originalAlpha = nil
	end

	return self
end

function M.FrameConfigMixin:ResetBorder()
	if self.originalBorder ~= nil and self:CanRun(self.ResetBorder) then
		B:RemoveBorder(self.frame)
		self.originalBorder = nil
	end

	return self
end

function M.FrameConfigMixin:ResetDrawLayers()
	if self.disabledDrawLayers and self:CanRun(self.ResetDrawLayers) then
		for drawLayer in next, self.disabledDrawLayers do
			self.frame:EnableDrawLayer(drawLayer)
		end

		self.disabledDrawLayers = nil
	end

	if self.originalDrawLayer and self:CanRun(self.ResetDrawLayers) then
		self.frame:SetDrawLayer(self.originalDrawLayer)
		self.originalDrawLayer = nil
	end

	return self
end

function M.FrameConfigMixin:ResetFont()
	if self.originalFont and self:CanRun(self.ResetFont) then
		self.frame:SetFont(unpack(self.originalFont))
		self.originalFont = nil
	end

	return self
end

function M.FrameConfigMixin:ResetFontObject()
	if self.originalFontObject and self:CanRun(self.ResetFontObject) then
		self.frame:SetFontObject(self.originalFontObject)
		self.originalFontObject = nil
	end

	return self
end

function M.FrameConfigMixin:ResetMaskTextures()
	if self.originalMaskTextures and self:CanRun(self.ResetMaskTextures) then
		for mask in next, self.originalMaskTextures do
			self.frame:AddMaskTexture(mask)
		end

		self.originalMaskTexture = nil
	end

	return self
end

function M.FrameConfigMixin:ResetParent()
	if self.originalParent and self:CanRun(self.ResetParent) then
		self.frame:SetParent(self.originalParent)
		self.originalParent = nil
	end

	return self
end

function M.FrameConfigMixin:ResetPoints()
	if self.originalPoints and self:CanRun(self.ResetPoints) then
		M:Unhook(self.frame, "ClearAllPoints")
		M:Unhook(self.frame, "SetAllPoints")
		M:Unhook(self.frame, "SetPoint")

		self.frame:ClearAllPoints()

		for _, point in next, self.originalPoints do
			self.frame:SetPoint(unpack(point))
		end

		self.originalPoints = nil
		self.overridePoints = nil
	end

	return self
end

function M.FrameConfigMixin:ResetScale()
	if self.originalScale and self:CanRun(self.ResetScale) then
		self.frame:SetScale(self.originalScale)
		self.originalScale = nil
	end

	return self
end

function M.FrameConfigMixin:ResetShown()
	if self.originalIsShown ~= nil and self:CanRun(self.ResetShown) then
		M:Unhook(self.frame, "OnHide")
		M:Unhook(self.frame, "OnShow")
		self.frame[self.originalIsShown and "Show" or "Hide"](self.frame)
		self.originalIsShown = nil
	end

	return self
end

function M.FrameConfigMixin:ResetTexCoord()
	if self.originalTexCoord and self:CanRun(self.ResetTexCoord) then
		self.frame:SetTexCoord(unpack(self.originalTexCoord))
		self.originalTexCoord = nil
	end

	return self
end

function M.FrameConfigMixin:ResetTexture()
	if self.originalTexture and self:CanRun(self.ResetTexture) then
		self.frame:SetTexture(unpack(self.originalTexture))
		self.originalTexture = nil
	end

	return self
end

M.FrameConfigMixin.ShouldRunSecure = {
	[M.FrameConfigMixin.UpdatePoints] = true,
	[M.FrameConfigMixin.ResetParent] = true,
	[M.FrameConfigMixin.ResetPoints] = true,
	[M.FrameConfigMixin.ResetScale] = true,
	[M.FrameConfigMixin.ResetShown] = true,
	[M.FrameConfigMixin.RunSecure] = true,
	[M.FrameConfigMixin.SetParent] = true,
	[M.FrameConfigMixin.SetScale] = true,
	[M.FrameConfigMixin.SetShown] = true,
}

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:PLAYER_REGEN_ENABLED()
	self.CombatQueue:Process()
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

-- Use a hook to find newly created frames.
function M:CreateFrame(frameType, frameName, parentFrame, inheritsFrame)
	if frameName then
		self.FrameQueue:Process(frameName)
	end
end

-- Use a hook to hide a frame whenever it is shown.
function M:Frame_OnHide(frame)
	A:RunSecure({}, frame.Show, frame)
end

-- Use a hook to hide a frame whenever it is shown.
function M:Frame_OnShow(frame)
	A:RunSecure({}, frame.Hide, frame)
end
