local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database

local M = A:NewModule("Core Layout")

local AceDB = LibStub:GetLibrary("AceDB-3.0", true)

M.DEFAULT_PROFILE_NAME = "SixUI"

function M:OnInitialize()
	self.db = D:RegisterNamespace(self:GetName(), {char = {layout = C.Enum.Layout.Standard}})

	A:AddCommand("layout", function (layout)
		if not layout or layout == "" then
			A.PrintF("You must specify a layout name by typing '%s'.", A.Highlight("/sixui layout [name]"))
		else
			self:SetActiveLayout(layout)
		end
	end, "change SixUI's layout.")
end

function M:OnEnable()
	self:CreateHook("SecureHook", EditModeManagerFrame, "NotifyChatOfLayoutChange", "OnEditModeLayoutChanged")
	self:RegisterMessage("SIXUI_LAYOUT_CHANGED", "OnLayoutChanged")
	self:RegisterMessage("SIXUI_MODULE_UPDATE", "OnModuleUpdate")
	self:RegisterMessage("SIXUI_PROFILE_CHANGED", "OnProfileChanged")
end

function M:OnEditModeLayoutChanged()
	self:SetActiveLayout(self:GetEditModeLayout(), true)
end

function M:OnLayoutChanged()
	self:UpdateActiveProfile(true)
end

function M:OnModuleUpdate(event, module, shouldSetProfile)
	if shouldSetProfile then
		self:SetActiveProfile(module, true, self:GetProfileName())
	end
end

function M:OnProfileChanged()
	self:SetActiveLayout(self:GetActiveLayout(), true)
end

-------------------------------------------------------------------------------
-- Layout
-------------------------------------------------------------------------------

function M:GetActiveLayout()
	return D.profile.layout.activeLayout or C.Enum.Layout.Standard
end

function A:GetActiveLayout()
	return M:GetActiveLayout()
end

function M:GetEditModeLayout()
	return C.Enum.ProfileToLayout[EditModeManagerFrame:GetActiveLayoutInfo().layoutName]
end

function M:GetProfileName()
	local profileName = C.Enum.LayoutToProfile[self:GetActiveLayout()]

	if profileName == self.DEFAULT_PROFILE_NAME then
		return profileName
	else
		return profileName, self.DEFAULT_PROFILE_NAME
	end
end

function M:HasLayout(layout)
	return C.Enum.Layout[layout] ~= nil
end

function M:IsActiveLayout(layout)
	return self:GetActiveLayout() == layout
end

function M:SetActiveLayout(layout, silent)
	if not self:HasLayout(layout) then
		if not silent then
			A.PrintF("The layout '%s' wasn't recognised.", A.Highlight(layout))
		end
	elseif self:IsActiveLayout(layout) then
		if not silent then
			A.PrintF("The layout '%s' is already active.", A.Highlight(layout))
		end
	else
		D.profile.layout.activeLayout = layout
		self:SendMessage("SIXUI_LAYOUT_CHANGED", layout)
	end
end

-------------------------------------------------------------------------------
-- Profile
-------------------------------------------------------------------------------

-- Iterate over AceDB databases and SixUI modules using the SetupMixin interface.
function M.IterateSetup()
	return coroutine.wrap(function ()
		for db in next, AceDB.db_registry do
			if not D:IsDatabase(db) and not db.parent then
				coroutine.yield(Mixin({}, C.AceSetupMixin, {GetDatabase = function (self) return db end}))
			end
		end

		for _, moduleName in next, {"Blizzard EditMode", "Integration Gnosis", "Integration VuhDo"} do
			local module = A:GetModule(moduleName, true)
	
			if module and module:IsEnabled() then
				coroutine.yield(module)
			end
		end
	end)
end

function M:SetActiveProfile(setup, force, ...)
    local activeProfileName = setup:GetActiveProfile()

    if force or C.Enum.ProfileToLayout[activeProfileName] then
        for _, profileName in next, {...} do
            if setup:HasProfile(profileName) then
                if profileName ~= activeProfileName then
                    setup:SetActiveProfile(profileName)
                end

                return
            end
        end
    end
end

function M:UpdateActiveProfile(force)
	local profileName = {self:GetProfileName()}

	for setup in self.IterateSetup() do
		self:SetActiveProfile(setup, force, unpack(profileName))
	end
end
