local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database

local M = A:NewModule("Integration Gnosis"):WithDependency("Gnosis")

M.AnchorPoint = {
	Icon = {
		TOP = "TOPLEFT",
		RIGHT = "BOTTOMRIGHT",
		BOTTOM = "BOTTOMRIGHT",
		LEFT = "TOPLEFT",
	},
	Bar = {
		TOP = "BOTTOMRIGHT",
		RIGHT = "TOPLEFT",
		BOTTOM = "TOPLEFT",
		LEFT = "BOTTOMRIGHT",
	},
}

M.ColorKey = {
	colBar = "_colBar",
	colBarNI = "_colBarNI",
	colChanneled = "_colChanneled",
}

M.PlayerUnitId = {
	gcd = true,
	gcd2 = true,
	mirror = true,
	player = true,
	sm = true,
	smr = true,
	sr = true,
	vehicle = true,
}

function M.GetBarColor(bar)
	local unit = bar.conf.unit
	local color

	if M.config.useClassColors and M.PlayerUnitId[unit] then
		color = C.CLASS_TO_COLOR[C.PLAYER_CLASS]
	elseif M.config.useClassColors and UnitIsPlayer(unit) then
		color = C.CLASS_TO_COLOR[UnitClass(unit)]
	elseif M.config.useReactionColors then
		color = C.REACTION_TO_COLOR[UnitReaction(unit, "player")]
	end

	return color
end

function M.SetBarColor(bar, color)
	for key, resetKey in next, M.ColorKey do
		if not bar[resetKey] then
			bar[resetKey] = bar.conf[key]
			bar.conf[key] = {color:GetRGBA()}
		end
	end
end

function M.ResetBarColor(bar)
	for key, resetKey in next, M.ColorKey do
		if bar[resetKey] then
			bar.conf[key] = bar[resetKey]
			bar[resetKey] = nil
		end
	end
end

function M.IterateBars()
	return coroutine.wrap(function ()
		if Gnosis then
			for _, bars in next, {Gnosis.castbars or {}, Gnosis.unusedcastbars or {}} do
				for _, bar in next, bars do
					coroutine.yield(bar)
				end
			end
		end
	end)
end

function M:OnEnable()
	self:Update()
end

function M:Update()
	self:UpdateHooks()
	self:ResetAllBars()
	self:StyleAllBars()
end

function M:UpdateHooks()
	if self.config.skinBars then
		self:CreateHook("SecureHook", Gnosis, "AnchorBar", "Gnosis_AnchorBar")
		self:CreateHook("SecureHook", Gnosis, "SetBarParams", "Gnosis_SetBarParams")
		self:CreateHook("SecureHook", Gnosis, "SetBorderColor", "Gnosis_SetBorderColor")
	else
		self:Unhook(Gnosis, "AnchorBar")
		self:Unhook(Gnosis, "SetBarParams")
		self:Unhook(Gnosis, "SetBorderColor")
	end

	if self.config.useClassColors or self.config.useReactionColors then
		self:CreateHook("RawHook", Gnosis, "activeBarColors", "Gnosis_activeBarColors")
	else
		self:Unhook(Gnosis, "activeBarColors")
	end

	self:CreateHook("RawHook", Gnosis, "CheckForFirstStart", C.NOOP_HANDLER)
	self:CreateHook("RawHook", Gnosis, "CreateBarFrame", "Gnosis_CreateBarFrame")
	self:RegisterMessage("SIXUI_CONFIG_CHANGED", "OnConfigChanged")
end

function M:OnDisable()
	self:ResetAllBars()
end

function M:OnConfigChanged(event, module, config, key, ...)
	if module == self:GetName() then
		self:Update()
	end
end

-------------------------------------------------------------------------------
-- Skin
-------------------------------------------------------------------------------

function M:StyleAllBars()
	for bar in self.IterateBars() do
		self:StyleBar(bar)
	end
end

function M:ResetAllBars()
	for bar in self.IterateBars() do
		self:ResetBar(bar)
	end
end

function M:StyleBar(bar)
	if not bar or not bar.conf or not self.config.skinBars then return end

	if not bar._bar then
		bar._bar = CreateFrame("Frame", nil, bar)
	end

	if not bar._icon then
		bar._icon = CreateFrame("Frame", nil, bar)
	end

	local _bar = bar._bar
	local _icon = bar._icon
	local iconSide = bar.conf.iconside

	if iconSide == "NONE" then
		_bar:SetAllPoints(bar.bar)
		_icon:Hide()
	elseif bar.conf.bIconUnlocked then
		_bar:SetAllPoints(bar.bar)
		_icon:SetAllPoints(bar.icon)
		_icon:Show()
	else
		_bar:ClearAllPoints()
		_bar:SetPoint(self.AnchorPoint.Bar[iconSide], bar.bar)
		_bar:SetPoint(self.AnchorPoint.Icon[iconSide], bar.icon)
		_icon:Hide()
	end

	B:CreateBorder(_bar, C.BORDER_PADDING)
	B:CreateBorder(_icon, C.BORDER_PADDING)
end

function M:UpdateBarAnchor(bar, barName, barConfig)
	if barConfig.anchortype ~= 2 then return end
	if not _G[barConfig.anchorframe] then return end
	if barConfig.anchorframe == barName then return end
	local point, relativeTo, relativePoint, offsetX, offsetY = bar:GetPoint()
	local scale = UIParent:GetEffectiveScale()
	bar:ClearAllPoints()
	bar:SetPoint(point, relativeTo, relativePoint, offsetX * scale, offsetY * scale)
end

function M:ResetBar(bar)
	B:RemoveBorder(bar._bar)
	B:RemoveBorder(bar._icon)
end

-- Set a bar's color to its unit's class or reaction color.
function M:Gnosis_activeBarColors(Gnosis, bar)
	local color = self.GetBarColor(bar)

	if color then
		self.SetBarColor(bar, color)
		self.hooks[Gnosis].activeBarColors(Gnosis, bar)
		self.ResetBarColor(bar)
	else
		self.hooks[Gnosis].activeBarColors(Gnosis, bar)
	end
end

-- Prevent bars from anchoring relative to the screen's scale.
function M:Gnosis_AnchorBar(Gnosis, name)
	self:UpdateBarAnchor(Gnosis.castbars[name], name, Gnosis.s.cbconf[name])
end

-- Coerce Gnosis to reuse a bar with the same name.
function M:Gnosis_CreateBarFrame(Gnosis, name, ...)
	for index, bar in next, Gnosis.unusedcastbars or {} do
		if bar and bar:GetName() == name then
			table.insert(Gnosis.unusedcastbars, table.remove(Gnosis.unusedcastbars, index))
		end
	end

	local bar = self.hooks[Gnosis].CreateBarFrame(Gnosis, name, ...)

	self:StyleBar(bar)

	return bar
end

-- Style a bar when 'bIconUnlocked' or 'iconside' might have changed.
function M:Gnosis_SetBarParams(Gnosis, barName, cfgtab, bartab)
	self:StyleBar(Gnosis.castbars[barName])
end

-- Synchronize a bar's border color with its SixUI border.
function M:Gnosis_SetBorderColor(Gnosis, bar, colBorder, colBarBg, bHideIconBorder)
	if A.table.compare(colBorder, bar.conf.colBorder) then
		colBorder = {}
	end

	if bar._bar then
		bar._bar:GetBorder():SetColor(unpack(colBorder))
	end

	if bar._icon then
		bar._icon:GetBorder():SetColor(unpack(colBorder))
	end
end

-------------------------------------------------------------------------------
-- Setup
-------------------------------------------------------------------------------

Mixin(M, C.SetupMixin)

function M.CreateGnosisOptions()
	Gnosis:GetNextTableIndex(1)
	Gnosis:GetNextTableIndexInner(1)
	Gnosis:CreateOptions()
end

function M.RemoveAllCastbars()
	for key in next, Gnosis.s.cbconf do
		Gnosis:RemoveCastbar(key)
	end
end

function M:CreateProfileFromString(profileName, profileString, successFn)
	self:SaveProfile(self.BACKUP_PROFILE_NAME)
	self.RemoveAllCastbars()
	self:ImportCastbars(profileString)
	self:SaveProfile(profileName)
	self:SetActiveProfile(self.BACKUP_PROFILE_NAME)
	successFn()
end

function M:DeleteProfile(profileName)
	GnosisConfigs[profileName] = nil
	self.CreateGnosisOptions()
end

-- Gnosis doesn't have a concept of active profiles.
function M:GetActiveProfile()
	return ""
end

function M:HasProfile(profileName)
	return GnosisConfigs[profileName] ~= nil
end

function M:ImportCastbars(profileString)
	self:CreateHook("RawHook", Gnosis.dialog, "Spawn", "OnDialogSpawn")
	Gnosis:ImportBarsFromStr(profileString)
	self:Unhook(Gnosis.dialog, "Spawn")
end

-- Gnosis doesn't have a concept of active profiles.
function M:IsActiveProfile(profileName)
	return false
end

-- Use a hook to click the 'Import' button when a Gnosis dialog spawns.
function M:OnDialogSpawn(LibDialog, ...)
	local dialog = self.hooks[LibDialog].Spawn(LibDialog, ...)

	if dialog then
		dialog.buttons[1]:Click()
	end
end

function M:SaveProfile(profileName)
	self:DeleteProfile(profileName)
	self.CreateGnosisOptions()
	Gnosis:OptSaveNewConfig(profileName, false, true, false, false)
end

function M:SetActiveProfile(profileName)
	return Gnosis:LoadConfig(profileName, false, true, false, false)
end
