local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database

local M = A:NewModule("Integration ShadowedUnitFrames"):WithDependency("ShadowedUnitFrames")

function M.IterateAuras(frame)
	local auras = frame and frame.auras
	local buffs = auras and auras.buffs
	local debuffs = auras and auras.debuffs

	return coroutine.wrap(function ()
		for i = 1, (buffs and buffs.totalAuras or 0) do
			coroutine.yield(buffs.buttons[i])
		end

		for i = 1, (debuffs and debuffs.totalAuras or 0) do
			coroutine.yield(debuffs.buttons[i])
		end
	end)
end

function M.IterateFrames()
	return coroutine.wrap(function ()
		for frame in next, ShadowUF.Units.frameList do
			if frame then
				coroutine.yield(frame)
			end
		end
	end)
end

function M:OnEnable()
	self:Update()
end

function M:OnDisable()
	self:ResetAllFrames()
end

function M:OnClutterStateChanged(event, state)
	if ShadowUF.tagFunc then
		ShadowUF.Tags:Reload()
	end
end

function M:OnConfigChanged(event, module, config, key, ...)
	if module == self:GetName() then
		self:Update()
	end
end

function M:Update()
	self:UpdateHooks()
	self:ResetAllFrames()
	self:SkinAllFrames()
end

function M:UpdateHooks()
	if self.config.skinAuras and self.config.skinUnitFrames then
		self:CreateHook("RawHook", ShadowUF.modules.auras, "Update", "Auras_Update")
	else
		self:Unhook(ShadowUF.modules.auras, "Update")
	end

	if self.config.skinUnitFrames then
		self:CreateHook("RawHook", ShadowUF.modules.highlight, "Update", "Highlight_Update")
		self:CreateHook("RawHook", ShadowUF.modules.units, "CreateUnit", "Units_CreateUnit")
	else
		self:Unhook(ShadowUF.modules.highlight, "Update")
		self:Unhook(ShadowUF.modules.units, "CreateUnit")
	end

	if self.config.useClutterKey then
		self:RegisterMessage("SIXUI_CLUTTER_STATE_CHANGED", "OnClutterStateChanged")
	else
		self:UnregisterMessage("SIXUI_CLUTTER_STATE_CHANGED")
	end

	if self.config.useInstantFading then
		self:CreateHook("RawHook", ShadowUF.modules.fader, "Update", "Fader_Update")
	else
		self:Unhook(ShadowUF.modules.fader, "Update")
	end

	self:RegisterMessage("SIXUI_CONFIG_CHANGED", "OnConfigChanged")
end

-------------------------------------------------------------------------------
-- Skin
-------------------------------------------------------------------------------

M.skinnedFrames = {}

function M.Aura_SetVertexColor(self, r, g, b, a)
	if (not self.GetBorder) then return end
	if (M.round(r) == 0.6) then r, g, b, a = nil, nil, nil, nil end
	self:GetBorder():SetColor(r, g, b, a)
end

function M.round(number)
	return math.floor(number * 10) / 10
end

function M:SkinAllFrames()
	for frame in self.IterateFrames() do
		self:SkinFrame(frame)
	end
end

function M:ResetAllFrames()
	for frame in next, self.skinnedFrames do
		self:ResetFrame(frame)
	end
end

function M:SkinFrame(frame)
	if not self.skinnedFrames[frame] and self.config.skinUnitFrames then
		self.skinnedFrames[frame] = true
		A:GetFrameConfig(frame):CreateBorder():DisableDrawLayer("BACKGROUND"):DisableDrawLayer("BORDER")
		self:SkinAllAuras(frame)
	end
end

function M:ResetFrame(frame)
	if self.skinnedFrames[frame] then
		self.skinnedFrames[frame] = nil
		A:GetFrameConfig(frame):ResetAll()
		self:ResetAllAuras(frame)
	end
end

function M:SkinAllAuras(frame)
	for button in self.IterateAuras(frame) do
		self:SkinAura(button)
	end
end

function M:ResetAllAuras(frame)
	for aura in self.IterateAuras(frame) do
		self:ResetAura(aura)
	end
end

function M:SkinAura(button)
	if (not button or not self.config.skinAuras) then return end
	local border = button.border
	B:CreateBorder(button):SetPadding(1)
	border.GetBorder = button.GetBorder

	if self:CreateHook("SecureHook", border, "SetVertexColor", self.Aura_SetVertexColor) then
		self.Aura_SetVertexColor(border, border:GetVertexColor())
	end
end

function M:ResetAura(aura)
	if (not aura) then return end
	self:Unhook(aura.border, "SetVertexColor")
	aura.border.GetBorder = nil
	B:RemoveBorder(aura)
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

-- Skin newly created aura buttons.
function M:Auras_Update(object, frame, ...)
	self.hooks[ShadowUF.modules.auras].Update(object, frame, ...)
	self:SkinAllAuras(frame)
end

-- Decrease the fade time for unit frames.
function M:Fader_Update(object, frame, ...)
	self.hooks[ShadowUF.modules.fader].Update(object, frame, ...)
	frame.fader.fadeTime = 0
end

-- Synchronize SixUI's border color with Shadowed Unit Frames.
function M:Highlight_Update(object, frame, ...)
	self.hooks[ShadowUF.modules.highlight].Update(object, frame, ...)

	if frame and frame.GetBorder and frame.highlight then
		if frame.highlight:IsShown() then
			frame:GetBorder():SetColor(frame.highlight.top:GetVertexColor())
		else
			frame:GetBorder():SetColor()
		end
	end
end

-- Skin newly created unit frames.
function M:Units_CreateUnit(...)
	local frame = self.hooks[ShadowUF.modules.units].CreateUnit(...)
	self:SkinFrame(frame)
	return frame
end

-------------------------------------------------------------------------------
-- Setup
-------------------------------------------------------------------------------

Mixin(M, C.AceSetupMixin)

function M.LoadShadowedUFOptions()
	if not C_AddOns.IsAddOnLoaded("ShadowedUF_Options") then
		C_AddOns.LoadAddOn("ShadowedUF_Options")
	end

	if not ShadowUF.Config then
		error("ShadowedUF_Options is missing")
	end
end

function M.ProfileStringToLayout(profileString)
	local layout, err = loadstring(string.format([[return %s]], profileString))

	if err then
		error(string.format(L["Failed to import layout, error:|n|n%s"], err))
	end

	layout = layout()
	layout.units = A.table.filter(layout.units or {}, function (_, u) return type(ShadowUF.defaults.profile.units[u]) == "table" end)
	layout.positions = layout.positions or CopyTable(ShadowUF.db.profile.positions)
	layout.visibility = layout.visibility or CopyTable(ShadowUF.db.profile.visibility)
	return layout
end

function M:CreateProfileFromString(profileName, profileString, successFn)
	local activeProfile = self:GetActiveProfile()
	self:ImportLayout(profileName, profileString)
	self:SetActiveProfile(activeProfile)
	successFn()
end

function M:GetDatabase()
	return ShadowUF.db
end

-- ShadowedUnitFrames doesn't provide an API for importing profiles, so this mimics the lambda function they use during the import process.
function M:ImportLayout(profileName, profileString)
	self.LoadShadowedUFOptions()

	local layout = self.ProfileStringToLayout(profileString)

	ShadowUF.layoutImporting = true
	ShadowUF.db:SetProfile(profileName)
	ShadowUF.db:ResetProfile()
	ShadowUF.layoutImporting = nil
	ShadowUF:LoadDefaultLayout()

	for key, data in pairs(layout) do
		ShadowUF.db.profile[key] = type(data) == "table" and CopyTable(data) or data
	end
end
