
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: coroutine, next, pairs, Skada

local math = math

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Integration Skada"):WithDependency("Skada")

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

M.COLUMN_NAME_TO_CONFIG_KEY = {
	Damage = "columnTotal",
	DPS = "columnThroughput",
	Healing = "columnTotal",
	HPS = "columnThroughput",
	Percent = "columnPercent",
}

M.TITLE_BACKDROP = {
	bgFile = C.LIGHT_TEXTURE,
	tile = false,
	tileSize = 0,
	edgeSize = 0,
}

M.ColumnType = C.Enum.SkadaColumnType

M.UpdateType = {All = 1, Automatic = 2}

function M.ShouldShowColumn(columnType, isClutterKeyDown)
	return columnType == M.ColumnType.Always
		or (columnType == M.ColumnType.Clutter and isClutterKeyDown)
		or (columnType == M.ColumnType.NoClutter and not isClutterKeyDown)
end

function M.IterateColumns()
	return coroutine.wrap(function ()
		for _, module in next, {Skada.modules.Damage, Skada.modules.Healing} do
			local columns = module and module.metadata and module.metadata.columns or {}

			for columnName in next, columns do
				coroutine.yield(columnName, columns)
			end
		end
	end)
end

function M.IterateWindows()
	return coroutine.wrap(function ()
		for _, window in next, Skada:GetWindows() do
			if window.bargroup then
				coroutine.yield(window)
			end
		end
	end)
end

function M:OnEnable()
	self:Update()
end

function M:OnDisable()
	self:UpdateCallbacks(true)
	self:UpdateColumns(self.config.useClutterKey and M.UpdateType.All or M.UpdateType.Automatic)
	self:ResetAllWindows()
	self:UpdateSkada()
end

function M:Update()
	self:UpdateCallbacks()
	self:UpdateColumns(M.UpdateType.Automatic)
	self:ResetAllWindows()
	self:StyleAllWindows()
end

function M:UpdateCallbacks(disable)
	self:UnhookAll()
	self:UnregisterAllEvents()
	self:UnregisterAllMessages()
	if disable then return end

	if self.config.optimizeBarHeight then
		self:CreateHook("SecureHook", Skada, "CreateWindow", "StyleAllWindows")
		self:CreateHook("RawHook", Skada.modules.BarDisplay, "Destroy", "BarDisplay_Destroy")
		self:CreateHook("SecureHook", Skada.modules.BarDisplay, "WindowResized", "UpdateSkadaWithDelay")
	end

	if self.config.optimizeBarSpacing then
		self:CreateHook("SecureHook", Skada, "CreateWindow", "StyleAllWindows")
		self:CreateHook("RawHook", Skada.modules.BarDisplay, "Destroy", "BarDisplay_Destroy")
		self:RegisterEvent("UI_SCALE_CHANGED", "UpdateSkadaWithDelay")
	end

	if self.config.skinWindows then
		self:CreateHook("SecureHook", Skada, "CreateWindow", "StyleAllWindows")
		self:CreateHook("RawHook", Skada.modules.BarDisplay, "Destroy", "BarDisplay_Destroy")
	end

	if self.config.useClutterKey then
		self:CreateHook("RawHook", Skada, "Report", "Skada_Report")
		self:RegisterMessage("SIXUI_CLUTTER_STATE_CHANGED")
	end

	self:RegisterMessage("SIXUI_CONFIG_CHANGED")
	self:RegisterMessage("SIXUI_LAYOUT_CHANGED", "Update")
	self:RegisterMessage("SIXUI_PROFILE_CHANGED", "Update")
end

function M:UpdateColumns(updateType)
	if (self.config.skinWindows and self.config.useClutterKey) or updateType == M.UpdateType.All then
		for columnName, columns in self.IterateColumns() do
			local configKey = self.COLUMN_NAME_TO_CONFIG_KEY[columnName]
			local columnType = self.config[configKey] or M.ColumnType.Never

			columns[columnName] = updateType == M.UpdateType.All or self.ShouldShowColumn(columnType, self.isClutterKeyDown)
		end

		Skada:UpdateDisplay(true)
	end
end

function M:UpdateSkada()
	Skada:ApplySettings()
end

function M:UpdateSkadaWithDelay()
	self:CancelTimer(self.updateSkadaTimer)
	self.updateSkadaTimer = self:ScheduleTimer("UpdateSkada", 0.01)
end

function M:StyleAllWindows()
	for window in self.IterateWindows() do
		self:StyleWindow(window)
	end
end

function M:ResetAllWindows()
	for window in self.IterateWindows() do
		self:ResetWindow(window)
	end
end

function M:StyleWindow(window)
	if self.config.skinWindows then
		A:GetFrameConfig(window.bargroup):DisableDrawLayer("BACKGROUND")
		A:GetFrameConfig(window.bargroup.borderFrame):CreateBorder(C.BORDER_PADDING):DisableDrawLayer("BORDER")
		A:GetFrameConfig(window.bargroup.button.borderFrame):DisableDrawLayer("BORDER")
		window.bargroup.button:SetBackdrop(nil)
	end

	self:OverrideDatabase(window)
end

function M:ResetWindow(window)
	A:GetFrameConfig(window.bargroup):ResetAll()
	A:GetFrameConfig(window.bargroup.borderFrame):ResetAll()
	A:GetFrameConfig(window.bargroup.button.borderFrame):ResetAll()
	window.bargroup.button:SetBackdrop(nil)
	self:OverrideDatabase(window, true)
end

function M:OverrideDatabase(window, disable)
	local disabled = not window.db.original
	local hasChanges = disable ~= disabled
	if not hasChanges then return end
	local db = window.db.original or window.db

	window.db = disable and db or A.CreateProxyTable(db, {
		background = function ()
			return self.config.skinWindows
				and {borderthickness = 0}
				or nil
		end,
		barheight = function ()
			return self.config.optimizeBarHeight
				and A.FormatDecimal(db.background.height / math.floor(db.background.height / db.barheight) - window.db.barspacing)
				or nil
		end,
		barspacing = function ()
			return self.config.optimizeBarSpacing
				and A.GetPixelSize(db.barspacing, window.bargroup)
				or nil
		end,
		bartexturepath = function ()
			return self.config.skinWindows and self.config.overrideTextures
				and C.LIGHT_TEXTURE_PATH
				or nil
		end,
		title = function ()
			return self.config.skinWindows and self.config.overrideTextures
				and {texture = C.LIGHT_TEXTURE_NAME}
				or nil
		end,
	})

	self:UpdateSkadaWithDelay()
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:SIXUI_CLUTTER_STATE_CHANGED(event, state)
	self.isClutterKeyDown = state
	self:UpdateColumns(M.UpdateType.Automatic)
end

function M:SIXUI_CONFIG_CHANGED(event, module, config, key, ...)
	if module == self:GetName() then
		if key == "skinWindows" then
			self:Update()
		elseif key == "columnPercent" or key  == "columnThroughput" or key == "columnTotal" or key == "useClutterKey" then
			self:UpdateColumns(self.config.useClutterKey and M.UpdateType.Automatic or M.UpdateType.All)
		else
			self:UpdateSkadaWithDelay()
		end
	end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function M:BarDisplay_Destroy(BarDisplay, window)
	self:ResetWindow(window)
	self.hooks[BarDisplay].Destroy(BarDisplay, window)
end

function M:Skada_Report(Skada, ...)
	self:UpdateColumns(M.UpdateType.All)
	self.hooks[Skada].Report(Skada, ...)
	self:UpdateColumns(M.UpdateType.Automatic)
end
