
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: BigTipDB, ChatFrame1, GameTooltip, GameTooltipDefaultContainer, math, NamePlateTooltip, SixUI_TooltipAnchor, table, TinyTooltip, type

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Integration TinyTooltip"):WithDependency("TinyTooltip")
local LibEvent = LibStub:GetLibrary("LibEvent.7000", true)
if not LibEvent then return end

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

-- Check if the default position of the game tooltip container has been moved.
function M.HasGameTooltipDefaultContainerMoved()
	local point, relativeTo, relativePoint, x, y = GameTooltipDefaultContainer:GetPoint()

	return point ~= "TOPLEFT"
		or relativeTo ~= ChatFrame1
		or relativePoint ~= "TOPRIGHT"
		or math.modf(x) ~= 30
		or math.modf(y) ~= 10
end

function M.IterateTooltips(filter)
	local tooltips = TinyTooltip and TinyTooltip.tooltips or {}

	return coroutine.wrap(function ()
		for tooltipIndex, tooltip in next, tooltips do
			if tooltip and (not filter or filter == tooltip) then
				coroutine.yield(tooltip, tooltipIndex)
			end
		end
	end)
end

function M:OnInitialize()
	self:RemoveTooltip(NamePlateTooltip)
end

function M:OnEnable()
	self:RemoveTooltip(NamePlateTooltip)
	self:UpdateHooks()
	self:UpdateTinyTooltip()
	self:StyleGameTooltip()
	self:StyleAllTooltips()
end

function M:OnDisable()
	self:UpdateHooks(true)
	self:UpdateTinyTooltip()
	self:ResetGameTooltip()
	self:ResetAllTooltips()
end

-- Remove a tooltip from the list of tooltips in TinyTooltip, if it exists.
function M:RemoveTooltip(tooltip)
	local _, tooltipIndex = self.IterateTooltips(tooltip)()
	if not tooltipIndex then return end
	table.remove(TinyTooltip.tooltips, tooltipIndex)
end

M:RemoveTooltip(NamePlateTooltip)

function M:UpdateHooks(disable)
	local hookFunc = disable and LibEvent.detachTrigger or LibEvent.attachTrigger

	for event, func in next, not hookFunc and {} or {
		["tooltip:hide"] = M.OnTooltipHide,
		["tooltip:show"] = M.OnTooltipShow,
		["tooltip.anchor.cursor"] = M.OnTooltipAnchor,
		["tooltip.anchor.cursor.right"] = M.OnTooltipAnchor,
		["tooltip.anchor.static"] = M.OnTooltipAnchor,
		["tooltip.style.background"] = M.OnTooltipStyleBackground,
		["tooltip.style.border.color"] = M.OnTooltipStyleBorderColor,
		["tooltip.style.font.body"] = M.OnTooltipStyleFontBody,
		["tooltip.style.font.header"] = M.OnTooltipStyleFontHeader,
		["tooltip.style.mask"] = M.OnTooltipStyleMask,
		["tooltip:unit"] = M.OnTooltipUnit,
	} do
		hookFunc(LibEvent, event, func)
	end

	if disable then
		self:UnhookAll()
	elseif type(TinyTooltip) == "table" and type(TinyTooltip.MergeVariable) == "function" then
		self:CreateHook("RawHook", TinyTooltip, "MergeVariable", "TinyTooltip_MergeVariable")
	end
end

function M:UpdateTinyTooltip()
	LibEvent:trigger("TINYTOOLTIP_GENERAL_INIT")
end

function M:StyleGameTooltip()
	A:GetFrameConfig(GameTooltip):DisableDrawLayer("BORDER"):DisableDrawLayer("BACKGROUND")
	A:GetFrameConfig(GameTooltip.NineSlice):DisableDrawLayer("BORDER"):DisableDrawLayer("BACKGROUND")
	A:GetFrameConfig("GameTooltipStatusBar"):SetShown(false)
end

function M:ResetGameTooltip()
	A:GetFrameConfig(GameTooltip):ResetAll()
	A:GetFrameConfig(GameTooltip.NineSlice):ResetAll()
	A:GetFrameConfig("GameTooltipStatusBar"):ResetAll()
end

function M:StyleAllTooltips()
	for tooltip in self.IterateTooltips() do
		if tooltip.style and tooltip:IsShown() then
			self.OnTooltipHide("", tooltip)
		end
	end
end

function M:ResetAllTooltips()
	for tooltip in self.IterateTooltips() do
		if tooltip.style then
			self.OnTooltipHide("", tooltip)
		end
	end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function M.OnTooltipAnchor(_, tooltip)
	local hasMoved = M.HasGameTooltipDefaultContainerMoved()
	local relativeTo = hasMoved and GameTooltipDefaultContainer or SixUI_TooltipAnchor
	local point = hasMoved and A.GetScreenCorner(relativeTo) or "TOPLEFT"
	tooltip:ClearAllPoints()
	tooltip:SetPoint(point, relativeTo)
end

function M.OnTooltipHide(_, tooltip)
	B:RemoveBorder(tooltip.style)
end

function M.OnTooltipShow(_, tooltip)
	B:CreateBorder(tooltip.style)
end

function M.OnTooltipStyleBackground(_, tooltip, r, g, b, a)
	tooltip.style:SetBackdropColor(0, 0, 0, 0)
end

function M.OnTooltipStyleBorderColor(_, tooltip, r, g, b, a)
	local border = B:GetBorder(tooltip.style)
	local defaultColor = TinyTooltip.db.general.borderColor

	if border then
		if defaultColor[1] ~= r or defaultColor[1] ~= g or defaultColor[3] ~= b or defaultColor[4] ~= a then
			border:SetColor(r, g, b, a)
		else
			border:SetColor()
		end
	end

	tooltip.style:SetBackdropBorderColor(0, 0, 0, 0)
end

function M.OnTooltipStyleFontHeader()
	local _, size, flags = GameTooltipHeaderText:GetFont()
	GameTooltipHeaderText:SetFont(C.TEXT_FONT_PATH, size, flags)
end

function M.OnTooltipStyleFontBody()
	local _, size, flags = GameTooltipText:GetFont()
	GameTooltipText:SetFont(C.TEXT_FONT_PATH, size, flags)
end

function M.OnTooltipStyleMask(_, tooltip)
	tooltip.style.mask:Hide()
end

function M.OnTooltipUnit(_, tooltip)
	tooltip.BigFactionIcon:Hide()
	tooltip.model:Hide()
end

function M:TinyTooltip_MergeVariable(TinyTooltip, src, dst)
	if src == TinyTooltip.db and dst == BigTipDB then
		src.general.mask = false
		src.item.coloredItemBorder = true
		src.unit.npc.showModel = false
		src.unit.player.elements.classIcon.enable = false
		src.unit.player.elements.factionBig.enable = false
		src.unit.player.elements.factionIcon.enable = false
		src.unit.player.elements.friendIcon.enable = false
		src.unit.player.elements.guildIndex.enable = true
		src.unit.player.elements.guildRealm.enable = false
		src.unit.player.elements.pvpIcon.enable = false
		src.unit.player.elements.realm.filter = "not samerealm"
		src.unit.player.elements.title.enable = false
		src.unit.player.showModel = false
	end

	return self.hooks[TinyTooltip].MergeVariable(TinyTooltip, src, dst)
end
