local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database

local M = A:NewModule("Integration VuhDo"):WithDependency("VuhDo")

function M.IteratePanels()
	return coroutine.wrap(function ()
		for panelIndex, panel in next, VUHDO_getAllActionPanels() do
			coroutine.yield(panel, panelIndex)
		end
	end)
end

function M.IterateButtons()
	return coroutine.wrap(function ()
		for _, panelIndex in M.IteratePanels() do
			for buttonIndex, button in next, VUHDO_getPanelButtons(panelIndex) or {} do
				coroutine.yield(button, panelIndex, buttonIndex)
			end
		end
	end)
end

function M.ReloadVuhDo()
	VUHDO_vuhdoInitLocalOverrides()

	if VUHDO_TIMERS and not VUHDO_isReloadPending() then
		VUHDO_TIMERS.RELOAD_UI = 0.1
	end
end

function M:OnEnable()
	if not VUHDO_isVariablesLoaded() then
		self:CancelTimer(self.onEnableTimer)
		self.onEnableTimer = self:ScheduleTimer("OnEnable", 0.25)
		return
	end

	self:CreateHook("RawHook", "VUHDO_determineRole")
	self:CreateHook("SecureHook", "VUHDO_getOrCreateActionPanel")
	self:CreateHook("SecureHook", "VUHDO_getOrCreateHealButton")
	self:CreateHook("RawHook", VUHDO_LibSharedMedia, "Fetch", "VUHDO_LibSharedMedia_Fetch")
	self:RegisterMessage("SIXUI_CONFIG_CHANGED", "OnConfigChanged")
	self:RegisterMessage("SIXUI_LAYOUT_CHANGED", "Update")
	self:RegisterMessage("SIXUI_PROFILE_CHANGED", "Update")
	self:Update()
end

function M:OnDisable()
	self:ResetAllPanels()
	self:ResetAllButtons()
	self:UnhookAll()
	self.ReloadVuhDo()
end

function M:Update()
	self:ResetAllPanels()
	self:SkinAllPanels()
	self:ResetAllButtons()
	self:SkinAllButtons()
	self:UpdatePanelPosition()
	self:UpdateButtonSize()
	self.ReloadVuhDo()
end

function M:UpdateWithDelay(...)
	self:CancelTimer(self.updateTimer)
	self.updateTimer = self:ScheduleTimer("Update", 0.25, ...)
end

function M:OnConfigChanged(event, module, config, key, ...)
	if module == "Core RaidFrameConfig" then
		if key == "width" or key == "height" then
			self:UpdateButtonSize()
		else
			self:UpdatePanelPosition()
		end
	elseif module == self:GetName() then
		self:UpdateWithDelay()
	end
end

-- Use a hook to find panels.
function M:VUHDO_getOrCreateActionPanel(panelNum)
	local panelName = format("Vd%s", panelNum)
	local panel = _G[panelName]

	self:SkinPanel(panel)

	if panelNum == self.PANEL_NUM then
		self:UpdatePanelPosition()
	end
end

-- Use a hook to find buttons.
function M:VUHDO_getOrCreateHealButton(buttonNum, panelNum)
	local buttonName = format("Vd%sH%s", panelNum, buttonNum)
	local button = _G[buttonName]

	self:SkinButton(button)

	if panelNum == self.PANEL_NUM then
		self:UpdateButtonSize()
	end
end

-------------------------------------------------------------------------------
-- Skin
-------------------------------------------------------------------------------

M.skinnedPanels = {}

M.skinnedButtons = {}

function M.GetIconFrameScale(panelNum)
	return max(0, min(100, VUHDO_PANEL_SETUP[panelNum].HOTS.size)) * 0.01
end

function M.GetOuterColor(outer)
	if outer:IsShown() then
		local r, g, b, a = outer:GetBackdropBorderColor()

		if (r ~= 0 or g ~= 0 or b ~= 0) and a ~= 0 then
			return r, g, b, a
		end
	end
end

function M:SkinAllPanels()
	for panel in self.IteratePanels() do
		self:SkinPanel(panel)
	end
end

function M:ResetAllPanels()
	for panel in next, self.skinnedPanels do
		self:ResetPanel(panel)
	end
end

function M:SkinPanel(panel)
	if not self.skinnedPanels[panel] and self.config.skinRaidFrames then
		self.skinnedPanels[panel] = panel
		panel:DisableDrawLayer("BACKGROUND")
		panel:DisableDrawLayer("BORDER")
	end
end

function M:ResetPanel(panel)
	self.skinnedPanels[panel] = nil
	panel:EnableDrawLayer("BACKGROUND")
	panel:EnableDrawLayer("BORDER")
end

function M:SkinAllButtons()
	for button in self.IterateButtons() do
		self:SkinButton(button)
	end
end

function M:ResetAllButtons()
	for button in self.IterateButtons() do
		self:ResetButton(button)
	end
end

function M:SkinButton(button)
	if not self.skinnedButtons[button] and self.config.skinRaidFrames then
		self.skinnedButtons[button] = button

		for _, index in next, {1, 2, 3, 4, 5, 9, 10, 40, 41, 42, 43, 44} do
			self:SkinIcon(VUHDO_getBarIcon(button, index))
		end

		for index = 40, 44 do
			self:SkinIconFrame(VUHDO_getBarIconFrame(button, index))
		end

		self:CreateBorder(button, C.BORDER_PADDING):SetBorderStrata("HIGH")
		self:SkinOuter(VUHDO_getPlayerTargetFrame(button), button.GetBorder)
	end
end

function M:ResetButton(button)
	if self.skinnedButtons[button] then
		self.skinnedButtons[button] = nil

		for _, index in next, {1, 2, 3, 4, 5, 9, 10, 40, 41, 42, 43, 44} do
			self:ResetIcon(VUHDO_getBarIcon(button, index))
		end

		for index = 40, 44 do
			self:ResetIconFrame(VUHDO_getBarIconFrame(button, index))
		end

		self:ResetOuter(VUHDO_getPlayerTargetFrame(button))
		self:RemoveBorder(button)
	end
end

function M:SkinIcon(icon)
	if icon and self.config.zoomIcons > 0 then
		icon:SetTexCoord(unpack(C.ZOOMED_TEX_COORDS[self.config.zoomIcons]))
		self:CreateHook("RawHook", icon, "SetTexCoord", C.NOOP_HANDLER, true)
	end
end

function M:ResetIcon(icon)
	if icon then
		self:Unhook(icon, "SetTexCoord")
		icon:SetTexCoord(0, 1, 0, 1)
	end
end

function M:SkinIconFrame(iconFrame)
	if iconFrame and self.config.resizeDebuffs then
		iconFrame:SetScale(self.GetIconFrameScale(self.PANEL_NUM))
		self:CreateHook("RawHook", iconFrame, "SetScale", C.NOOP_HANDLER, true)
	end
end

function M:ResetIconFrame(iconFrame)
	if iconFrame then
		self:Unhook(iconFrame, "SetScale")
		iconFrame:SetScale(VUHDO_CONFIG.CUSTOM_DEBUFF.scale * 0.7)
	end
end

function M:SkinOuter(outer, GetBorder)
	if outer then
		outer.GetBorder = GetBorder
		outer:DisableDrawLayer("BORDER")
		self:CreateHook("SecureHook", outer, "Hide", "UpdateOuter")
		self:CreateHook("SecureHook", outer, "Show", "UpdateOuter")
		self:UpdateOuter(outer)
	end
end

function M:UpdateOuter(outer)
	if outer.GetBorder then
		outer:GetBorder():SetRGB(self.GetOuterColor(outer))
	end
end

function M:ResetOuter(outer)
	if outer then
		self:Unhook(outer, "Hide")
		self:Unhook(outer, "Show")
		outer:EnableDrawLayer("BORDER")
		outer.GetBorder = nil
	end
end

-- Use a hook to coerce VuhDo into using SixUI's preferred status bar texture.
function M:VUHDO_LibSharedMedia_Fetch(library, mediaType, key, ...)
	return self.hooks[library].Fetch(library, mediaType, mediaType == "statusbar" and self.config.skinRaidFrames and self.config.overrideTextures and C.LIGHT_TEXTURE_NAME or key, ...)
end

-------------------------------------------------------------------------------
-- Raid Frame Config
-------------------------------------------------------------------------------

M.PANEL_NUM = 1

M.PANEL_NAME = format("Vd%s", M.PANEL_NUM)

-- Get the group size for the given panel number.
function M.GetGroupSize(panelNum)
	local groupSize = 0

	if VUHDO_isVariablesLoaded() then
		for modelIndex, modelId in next, VUHDO_getDynamicModelArray(panelNum) do
			groupSize = groupSize + #VUHDO_getGroupMembers(modelId, panelNum, modelIndex)
		end
	end

	return groupSize
end

-- Adjust the position of the primary panel.
function M:UpdatePanelPosition()
	if self.config.useRaidFrameConfig and self.config.controlPanelPosition then
		local panel = _G[self.PANEL_NAME]
		local config = A:GetRaidFrameConfig(self.GetGroupSize(self.PANEL_NUM))

		if panel and not config:IsEqualToPoint(panel:GetPoint()) then
			A:GetFrameConfig(self.PANEL_NAME):ClearAllPoints():SetPoint(config:GetPoint())
		end
	else
		A:GetFrameConfig(self.PANEL_NAME):ResetAll()
	end
end

-- Adjust the width and height of the primary panel's buttons.
function M:UpdateButtonSize()
	if self.config.useRaidFrameConfig and self.config.controlButtonSize then
		local panel = (VUHDO_PANEL_SETUP[self.PANEL_NUM] or {}).SCALING
		local config = A:GetRaidFrameConfig(self.GetGroupSize(self.PANEL_NUM))

		if panel and not config:IsEqualToBarSize(panel.barWidth, panel.barHeight) then
			panel.barWidth = config.width
			panel.barHeight = config.height
			self.ReloadVuhDo()
		end
	end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

-- Use a hook to override how VuhDo sorts classes.
function M:VUHDO_determineRole(unit)
	local classId = (VUHDO_RAID[unit] or {}).classId
	local role = self.hooks.VUHDO_determineRole(unit)

	-- VUHDO_ID_HUNTERS
	if self.config.overrideHunterRole and classId == 22 then
		return 62 -- VUHDO_ID_RANGED_DAMAGE
	end

	-- VUHDO_ID_PALADINS
	if self.config.overrideHolyPaladinRole and classId == 23 and role == 63 then
		return 61 -- VUHDO_ID_MELEE_DAMAGE
	end

	return role
end

-------------------------------------------------------------------------------
-- Setup
-------------------------------------------------------------------------------

Mixin(M, C.SetupMixin)

function M:CreateProfileFromString(profileName, profileString, successFn)
	self:DeleteProfile(profileName)
	local editBoxName = "SixUI_VuhDoEditBox"
	_G[editBoxName] = {GetText = function () return profileString end}
	VUHDO_profileImport(editBoxName)
	successFn()
end

function M:DeleteProfile(profileName)
	local profileIndex = self:GetProfile(profileName)

	if profileIndex then
		table.remove(VUHDO_PROFILES, profileIndex)
	end
end

function M:GetActiveProfile()
	return VUHDO_CONFIG["CURRENT_PROFILE"]
end

function M:GetProfile(profileName)
	-- VuhDo can sometimes prefix the profile name with the player's name.
	for _, name in ipairs({C.PLAYER_NAME .. ": " .. profileName, profileName}) do
		local profileIndex = A.table.findIndex(VUHDO_PROFILES, function (p) return p.NAME == name end)

		if profileIndex then
			return profileIndex, VUHDO_PROFILES[profileIndex]
		end
	end
end

function M:HasProfile(profileName)
	return self:GetProfile(profileName) ~= nil
end

function M:IsActiveProfile(profileName)
	return (self:GetProfile(profileName) or {}).NAME == self:GetActiveProfile()
end

function M:SetActiveProfile(profileName)
	local _, profile = self:GetProfile(profileName)

	if profile then
		(self.noInit and VUHDO_loadProfileNoInit or VUHDO_loadProfile)(profile.NAME)
		return true
	end
end
