local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database

local M = A:NewModule("Integration WeakAuras"):WithDependency("WeakAuras")

M.LINK_COLOR = CreateColor(0.5333, 0, 1, 1)

M.RegionType = {
	AuraBar = "aurabar",
	SubBorder = "subborder",
	Icon = "icon",
}

M.RegionTypeWithSkin = {
	[M.RegionType.AuraBar] = true,
	[M.RegionType.Icon] = true,
}

M.isAncestorIgnored = {}

M.regionRegistry = {}

function M.GetColorForBorder(border)
	if border and border:IsShown() then
		return M.NormalizeColor(border:GetBackdropBorderColor())
	else
		return nil
	end
end

function M.NormalizeColor(r, g, b, a)
	if (r ~= 0 or g ~= 0 or b ~= 0) and (r ~= 1 or g ~= 1 or b ~= 1) and a ~= 0 then
		return r, g, b
	end
end

function M:OnInitialize()
	self:CreateHooks()
end

function M:OnEnable()
	self:CreateHooks()
	self:RegisterRegions()
	self:RegisterMessage("SIXUI_CLUTTER_STATE_CHANGED", WeakAuras.ScanEvents)
	self:RegisterMessage("SIXUI_CONFIG_CHANGED", "OnConfigChanged")
	self:RegisterMessage("SIXUI_LAYOUT_CHANGED", WeakAuras.ScanEvents)
	self:RegisterMessage("SIXUI_PROFILE_CHANGED", "ReregisterRegions")
end

function M:OnDisable()
	self:UnregisterRegions()
end

function M:OnConfigChanged(event, module)
	if module == self:GetName() then
		self:ReregisterRegions()
	end
end

function M:CreateHooks()
	self:CreateHook("SecureHook", "SetItemRef")

	if WeakAuras and WeakAuras.GetRegion then
		self:CreateHook("RawHook", WeakAuras, "GetRegion", "WeakAuras_GetRegion")
	end
end

-- Iterate over known WeakAuras regions.
function M:IterateRegions()
	return coroutine.wrap(function ()
		for region in next, self.regionRegistry do
			if region and region.Show and region.Hide then
				coroutine.yield(region)
			end
		end
	end)
end

-- Iterate over sub-regions of a given type.
function M:IterateSubRegions(subRegions, subRegionType)
	local subRegionIndex = 0

	return coroutine.wrap(function ()
		for _, subRegion in next, subRegions do
			if not subRegionType or subRegionType == subRegion.type then
				subRegionIndex = subRegionIndex + 1
				coroutine.yield(subRegion, subRegionIndex)
			end
		end
	end)
end

-------------------------------------------------------------------------------
-- Skin
-------------------------------------------------------------------------------

-- Register the given region for updates and skin it.
function M:RegisterRegion(region)
	self:CreateHook("HookScript", region, "OnShow", "Region_OnShow")
	self:CreateHook("HookScript", region, "OnHide", "Region_OnHide")

	if region:IsShown() then
		self:Region_OnShow(region)
	else
		self:Region_OnHide(region)
	end
end

function M:RegisterRegions()
	for region in self:IterateRegions() do
		self:RegisterRegion(region)
	end
end

-- Unregister the given region from updates and reset it.
function M:UnregisterRegion(region)
	self:Unhook(region, "OnShow")
	self:Unhook(region, "OnHide")
	self:ResetRegion(region)
end

function M:UnregisterRegions()
	for region in self:IterateRegions() do
		self:UnregisterRegion(region)
	end
end

function M:ReregisterRegions()
	self:UnregisterRegions()
	self:RegisterRegions()
end

function M:SkinRegion(region)
	if (not self.RegionTypeWithSkin[region.regionType]) then return end
	if self:IsAncestorIgnored(region.id) then return end
	if (not self:IsEnabled()) then return end

	if region.regionType == self.RegionType.Icon and self.config.skinIcons then
		self:CreateBorder(region, C.BORDER_PADDING)
		self:SkinIcon(region.icon)
		self:SkinAllSubRegions(region.subRegions)
	end

	if region.regionType == self.RegionType.AuraBar and self.config.skinProgressBars then
		self:CreateBorder(region, C.BORDER_PADDING)
		self:SkinBar(region.bar)
		self:SkinAllSubRegions(region.subRegions)
	end
end

function M:ResetRegion(region)
	self:RemoveBorder(region)

	if region.regionType == self.RegionType.Icon then
		self:ResetIcon(region.icon)
	end

	if region.regionType == self.RegionType.AuraBar then
		self:ResetBar(region.bar)
	end

	self:ResetAllSubRegions(region.subRegions)
end

function M:SkinIcon(icon)
	if (not icon) then return end

	if self:CreateHook("RawHook", icon, "SetVertexColor", "Icon_SetVertexColor", true) then
		self:Icon_SetVertexColor(icon, icon:GetVertexColor())
	end
end

function M:ResetIcon(icon)
	if (not icon) then return end
	self:Unhook(icon, "SetVertexColor")
end

function M:SkinBar(bar)
	if (not bar) then return end
	self:SkinTexture(bar.bg, nil)
	self:SkinTexture(bar.fg, C.LIGHT_TEXTURE_PATH)
end

function M:ResetBar(bar)
	if (not bar) then return end
	self:ResetTexture(bar.bg)
	self:ResetTexture(bar.fg)
end

function M:SkinTexture(object, texture)
	if (not object) then return end
	A:GetFrameConfig(object):SetTexture(texture)
end

function M:ResetTexture(object)
	if (not object) then return end
	A:GetFrameConfig(object):ResetAll()
end

function M:SkinAllSubRegions(subRegions)
	if (not subRegions) then return end

	for border, borderIndex in self:IterateSubRegions(subRegions, self.RegionType.SubBorder) do
		border:SetAlpha(0)

		if borderIndex == 1 then
			self:CreateHook("RawHook", border, "SetBorderColor", "Border_SetBorderColor")
			self:CreateHook("RawHook", border, "SetVisible", "Border_SetVisible")
			self:Border_SetBorderColor(border, border:GetBackdropBorderColor())
		end
	end
end

function M:ResetAllSubRegions(subRegions)
	if (not subRegions) then return end

	for border in self:IterateSubRegions(subRegions, self.RegionType.SubBorder) do
		border:SetAlpha(1)
		self:Unhook(border, "SetBorderColor")
		self:Unhook(border, "SetVisible")
	end
end

-------------------------------------------------------------------------------
-- Ignore List
-------------------------------------------------------------------------------

function M:GetIgnoreList()
	local ignoreList = {}

	for target in next, self.config.blacklist do
		if self:IsIgnored(target) then
			ignoreList[target] = target
		end
	end

	return ignoreList
end

function M:Ignore(target)
	if not self:IsIgnored(target) then
		self.config.blacklist[target] = true
		table.wipe(self.isAncestorIgnored)
		self:ReregisterRegions()
	end
end

function M:IsIgnored(target)
	return self.config.blacklist[target] == true
end

function M:IsAncestorIgnored(target)
	local isOptionsOpen = WeakAuras.IsOptionsOpen()

	if self.isOptionsOpen ~= isOptionsOpen then
		self.isOptionsOpen = isOptionsOpen
		table.wipe(self.isAncestorIgnored)
	end

	if self.isAncestorIgnored[target] ~= nil then
		return self.isAncestorIgnored[target]
	end

	local isIgnored = self:IsIgnored(target)

	if not isIgnored then
		local data = WeakAuras.GetData(target)

		if data and data.parent then
			isIgnored = self:IsAncestorIgnored(data.parent)
		end
	end

	if not self.isOptionsOpen then
		self.isAncestorIgnored[target] = isIgnored
	end

	return isIgnored
end

function M:Unignore(target)
	if self:IsIgnored(target) then
		self.config.blacklist[target] = nil
		table.wipe(self.isAncestorIgnored)
		self:ReregisterRegions()
	end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function M:Border_SetBorderColor(region, ...)
	self.hooks[region].SetBorderColor(region, ...)
	local border = B:GetBorder(region:GetParent())
	if (not border) then return end
	border:SetColor(self.NormalizeColor(...))
end

function M:Border_SetVisible(region, visible)
	self.hooks[region].SetVisible(region, visible)

	local border = B:GetBorder(region:GetParent())
	if not border then return end

	if visible then
		border:SetColor(self.GetColorForBorder(region))
	else
		border:SetColor()
	end
end

function M:Icon_SetVertexColor(region, r, g, b, a)
	self.hooks[region].SetVertexColor(region, r, g, b,a)
	local border = B:GetBorder(region:GetParent())
	if (not border) then return end
	border[a == 0 and "Hide" or "Show"](border)
end

function M:Region_OnHide(region)
	self:ResetRegion(region)
end

function M:Region_OnShow(region)
	self:SkinRegion(region)
end

function M:SetItemRef(link)
	local type, value = string.split(":", link, 2)

	if type == "SixWA" then
		self:Import()
	end
end

function M:WeakAuras_GetRegion(...)
	local region = self.hooks[WeakAuras].GetRegion(...)

	if region and not self.regionRegistry[region] then
		self.regionRegistry[region] = true
		self:RegisterRegion(region)
	end

	return region
end

-------------------------------------------------------------------------------
-- Setup
-------------------------------------------------------------------------------

Mixin(M, C.SetupMixin)

function M:CreateProfileFromString(displayName, displayString, successFn)
	WeakAuras.Import(displayString, self:GetProfile(displayName), function (success) if success then successFn() end end)
end

function M:GetActiveProfile()
	return ""
end

function M:GetProfile(displayName)
	return WeakAuras.GetData(displayName)
end

function M:HasProfile(displayName)
	return self:GetProfile(displayName) ~= nil
end

function M:IsActiveProfile()
	return false
end

function M:SetActiveProfile()
	return
end
