local A, B, C, D = unpack(SixUI) -- AddOn, Border, Constants, Database

local M = A:GetModule("Core Config")

local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local CoreSetup = A:GetModule("Core Setup")

M.AddonProfileType = setmetatable({
	Blizzard_EditMode = "layout",
	WeakAuras = "display",
}, {__index = function () return "profile" end})

M.migrations = {}

function M.FormatKeyValueText(k, v)
	return format("%s %s", NORMAL_FONT_COLOR:WrapTextInColorCode(k .. ":"), HIGHLIGHT_FONT_COLOR:WrapTextInColorCode(v))
end

function M.GetMigrationAction(migration)
	if migration.description then
		return migration.description
	elseif migration.profileName then
		return format("Import %s '%s'", M.AddonProfileType[migration.addon], A.Highlight(migration.profileName))
	else
		return "Apply change"
	end
end

function M.GetMigrationLabel(migration)
	return migration.addon and format("%s: %s", migration.addon, M.GetMigrationAction(migration)) or M.GetMigrationAction(migration)
end

function M.GetMigrationStatus(migration)
	if migration:IsIgnored() then
		return LIGHTGRAY_FONT_COLOR:WrapTextInColorCode("Ignored")
	elseif migration:IsNew() then
		return "New"
	elseif migration:IsApplied() then
		return GREEN_FONT_COLOR:WrapTextInColorCode("Applied")
	else
		return RED_FONT_COLOR:WrapTextInColorCode("Outdated")
	end
end

function M.GenerateAtlasButtonArgs(atlas, iconSize)
	return {
		name = "",
		type = "execute",
		image = A.AtlasToFile(atlas),
		imageCoords = {A.AtlasToTexCoord(atlas)},
		imageWidth = iconSize,
		imageHeight = iconSize,
	}
end

function M.GenerateSetupOptionsTable()
	local optionsTable = {
		type = "group",
		name = "Setup",
		args = {
			description = {
				order = 0,
				name = "SixUI has one or more changes that are ready to be applied:",
				type = "description",
				fontSize = "medium",
			},
			updates = {
				order = 1,
				name = "Changes",
				type = "group",
				inline = true,
				args = {},
			},
			applyAllButton = {
				order = 2,
				name = "Apply All",
				type = "execute",
				func = function (info)
					for _, migration in next, info.arg do
						migration:Apply()
					end
				end,
				arg = M.migrations,
				disabled = function() return not A.table.any(M.migrations, function (m) return m:IsPending() end) end,
				width = "full",
			},
			ignoreAllButton = {
				order = 3,
				name = "Ignore All",
				type = "execute",
				func = function (info)
					for _, migration in next, info.arg do
						if not migration:IsApplied() then
							migration:SetIgnored(true)
						end
					end
				end,
				arg = M.migrations,
				disabled = function() return not A.table.any(M.migrations, function (m) return m:IsPending() end) end,
				width = "full",
			},
		},
	}

	for migrationIndex, migration in next, M.migrations do
		optionsTable.args.updates.args["" .. migrationIndex] = M.GenerateMigrationOptionsTable(migration, migrationIndex)
	end

	return optionsTable
end

function M.GenerateMigrationOptionsTable(migration, migrationIndex)
	local lines = {}
	local optionsTable = {
		order = migrationIndex,
		name = "",
		type = "group",
		inline = true,
		args = M.GenerateOrderedArgs(),
	}

	-- AddOn
	table.insert(lines, M.FormatKeyValueText("AddOn", migration.addon or migration.key == "Active Profile" and ALL or NOT_APPLICABLE))

	-- Action
	table.insert(lines, M.FormatKeyValueText("Action", M.GetMigrationAction(migration)))

	-- Status
	table.insert(lines, M.FormatKeyValueText("Status", M.GetMigrationStatus(migration)))

	if migrationIndex ~= 1 then
		optionsTable.args.divider = {
			name = "",
			type = "header",
		}
	end

	optionsTable.args.description = {
		name = table.concat(lines, "\n"),
		type = "description",
		fontSize = "medium",
		width = 1.6,
	}

	optionsTable.args.ignoreButton = Mixin({
		func = function (info) info.arg:SetIgnored(true) end,
		arg = migration,
		width = 0.2,
		disabled = not migration:IsPending(),
	}, M.GenerateAtlasButtonArgs("ui-lfg-declinemark-raid", 24))

	optionsTable.args.applyButton = Mixin({
		func = function (info) info.arg:Apply() end,
		arg = migration,
		width = 0.2,
		disabled = not migration:IsPending(),
	}, M.GenerateAtlasButtonArgs("ui-lfg-readymark-raid", 24))

	return optionsTable
end

function M.IsSetupFrameShown()
	return AceConfigDialog.OpenFrames["SixUI Setup"]
end

function M.ToggleSetupFrame(forceState)
	if forceState == false then
		AceConfigDialog:Close("SixUI Setup")
	elseif forceState or not AceConfigDialog:Close("SixUI Setup") then
		AceConfigDialog:Open("SixUI Setup")
	end
end

-- Update the migrations list and the toggle setup frame if any are pending.
function M:UpdateMigrations()
	if not M.IsSetupFrameShown() then
		table.wipe(self.migrations)
	end

	for migration in CoreSetup.IterateMigrations(true, false, false) do
		A.table.upsert(self.migrations, migration)
	end

	self:RefreshOptionsTable("SixUI Setup")
	self.ToggleSetupFrame(A.table.any(self.migrations, function (m) return m:IsPending() end))
end
