
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: CreateFrame, CreateFromMixins, Mixin, next, NORMAL_FONT_COLOR, SixUI, string, table, type, unpack, UIParent

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local MAJOR, MINOR = "SixTooltip-1.0", 1
if not LibStub then error(MAJOR .. " requires LibStub.") end
local L = LibStub:NewLibrary(MAJOR, MINOR)
if not L then return end
local LibQTip = LibStub("LibQTip-1.0")
if not LibQTip then return end

-------------------------------------------------------------------------------
-- Library
-------------------------------------------------------------------------------

L.CONFIG_DEFAULTS = {
	anchor = nil,
	autoHideAnchor = nil,
	cellMarginH = 6,
	cellMarginV = 3,
	columnJustification = {"LEFT"},
	enableScrolling = true,
	enableMouse = true,
	font = GameTooltipText,
	frameStrata = "DIALOG",
	headerFont = GameTooltipHeaderText,
	ignoreReleaseAllTooltips = false,
	key = "",
	numColumns = 1,
	point = nil,
	separatorHeight = 1,
	skipReleaseAllTooltips = false,
}

-- Acquire a tooltip, optionally releasing all other tooltips.
function L.AcquireTooltip(tooltip, config)
	if not config.skipReleaseAllTooltips then
		L.ReleaseAllTooltips(config.key)
	end

	if not tooltip or not LibQTip:IsAcquired(config.key) then
		tooltip = LibQTip:Acquire(config.key, config.numColumns, unpack(config.columnJustification))
		L.ConfigureTooltip(tooltip, config)
		L.StyleTooltip(tooltip, config)
	end

	tooltip:Clear()
	tooltip:SetCellMarginH(config.cellMarginH)
	tooltip:SetCellMarginV(config.cellMarginV)

	return tooltip
end

-- Configure a tooltip to be interactive.
function L.ConfigureTooltip(tooltip, config)
	tooltip:EnableMouse(config.enableMouse)
	tooltip:SetFrameStrata(config.frameStrata)

	if config.anchor then
		tooltip:SetAutoHideDelay(0.1, config.autoHideAnchor or config.anchor)
		tooltip:SmartAnchorTo(config.anchor)
	elseif config.point then
		if config.autoHideAnchor then
			tooltip:SetAutoHideDelay(0.1, config.autoHideAnchor)
		else
			tooltip:SetAutoHideDelay(0)
		end

		tooltip:ClearAllPoints()
		tooltip:SetPoint(unpack(config.point))
	end
end

-- Release all tooltips other than excludeKey.
function L.ReleaseAllTooltips(excludeKey)
	for key, tooltip in LibQTip:IterateTooltips() do
		if key ~= excludeKey and string.sub(key or "", 1 ,1) ~= "!" then
			tooltip:Release()
		end
	end
end

-- Style to a tooltip using SixUI.
function L.StyleTooltip(tooltip, config)
	if SixUI then
		SixUI[1]:GetFrameConfig(tooltip.NineSlice):CreateBorder():DisableDrawLayer("BACKGROUND"):DisableDrawLayer("BORDER")
		tooltip:SetFont(config.font)
		tooltip:SetHeaderFont(config.headerFont)
		tooltip.OnRelease = L.ResetTooltip
	end
end

-- Reset a tooltip that was styled using SixUI.
function L.ResetTooltip(tooltip)
	if SixUI then
		SixUI[1]:GetFrameConfig(tooltip.NineSlice):ResetAll()
		tooltip:SetFont(GameTooltipText)
		tooltip:SetHeaderFont(GameTooltipHeaderText)
		tooltip.OnRelease = nil
	end
end

-- Create a new instance of SixTooltip.
function L:CreateTooltip(config)
	return setmetatable(CreateFromMixins(self.SixTooltipMixin, {
		config = CreateFromMixins(self.CONFIG_DEFAULTS, config),
		dynamic = {},
	}), self.SixTooltipMetatable)
end

-------------------------------------------------------------------------------
-- SixTooltip
-------------------------------------------------------------------------------

L.SixTooltipMetatable = {
	__index = function (t, k)
		local tooltip = rawget(t, "tooltip")
		if not tooltip then return end
		local result = tooltip[k]
		if not result then return end
		return type(result) == "function"
			and function (_, ...) return result(tooltip, ...) end
			or result
	end,
}

L.SixTooltipMixin = {}

function L.SixTooltipMixin:AddDynamicLine(label, func, arg, isHeader)
	return self:AddSpannedLine(label, func, arg, isHeader, true)
end

function L.SixTooltipMixin:AddSpannedHeader(label, func, arg, skipSeparator)
	return self:AddSpannedLine(label, func, arg, true, false, skipSeparator)
end

function L.SixTooltipMixin:SetText(text, _, _, _, isDynamic)
	return self:AddSpannedLine(text, nil, nil, isDynamic, false)
end

function L.SixTooltipMixin:AddSpannedLine(label, func, arg, isHeader, isDynamic, skipSeparator)
	local line = isHeader and self.tooltip:AddHeader("") or self.tooltip:AddLine("")

	if isDynamic then
		table.insert(self.dynamic, {
			lineNum = line,
			colNum = 1,
			value = label,
			font = nil,
			justification = "LEFT",
			colSpan = self.config.numColumns,
		})
	else
		self.tooltip:SetCell(line, 1, label, nil, "LEFT", self.config.numColumns)
	end

	if type(func) == "function" then
		self.tooltip:SetLineScript(line, "OnMouseUp", func, arg)
	end

	if type(isHeader) == "boolean" and isHeader and not skipSeparator then
		self.tooltip:AddSeparator(self.config.separatorHeight)
	end

	return line
end

function L.SixTooltipMixin:InjectDynamicLines()
	local width = self.tooltip:GetWidth() - 24

	for _, line in next, self.dynamic do
		self.tooltip:SetCell(line.lineNum, line.colNum, line.value, line.font, line.justification, line.colSpan, nil, nil, nil, width)
	end

	table.wipe(self.dynamic)
end

function L.SixTooltipMixin:IsShown()
	return self.tooltip and LibQTip:IsAcquired(self.config.key)
end

function L.SixTooltipMixin:Refresh()
	if self:IsShown() then
		if type(self.OnRefresh) == "function" then
			self:OnRefresh()
		end

		self:Show()
	end
end

function L.SixTooltipMixin:Show(...)
	self.tooltip = L.AcquireTooltip(self.tooltip, self.config)

	if type(self.OnShow) == "function" then
		if self:OnShow(...) then
			self:InjectDynamicLines()
		end
	end

	for colNum = 1, self.config.numColumns do
		self.tooltip:SetColumnTextColor(colNum, NORMAL_FONT_COLOR:GetRGB())
	end

	if self.config.enableScrolling then
		self.tooltip:UpdateScrolling()
	end

	self.tooltip:Show()
end

function L.SixTooltipMixin:SmartAnchorTo(anchor, offsetX, offsetY)
	self.tooltip:ClearAllPoints()

	if anchor:GetLeft() > UIParent:GetWidth() / 2 then
		self.tooltip:SetPoint("RIGHT", anchor, "LEFT", -offsetX, offsetY)
	else
		self.tooltip:SetPoint("LEFT", anchor, "RIGHT", offsetX, offsetY)
	end

	if anchor:GetTop() > UIParent:GetHeight() / 2 then
		self.tooltip:SetPoint("TOP", anchor, "LEFT", -offsetX, offsetY)
	else
		self.tooltip:SetPoint("BOTTOM", anchor, "RIGHT", offsetX, offsetY)
	end
end

function L.SixTooltipMixin:UpdateConfig(config, ...)
	Mixin(self.config, config, ...)

	if self:IsShown() then
		L.ConfigureTooltip(self.tooltip, self.config)
		L.StyleTooltip(self.tooltip, self.config)
	end
end
